# Copyright DST Group. Licensed under the MIT license.
from ipaddress import IPv4Address

from csle_cyborg.shared.actions.msf_actions_folder.msf_scanner_folder.msf_scanner import MSFScanner
from csle_cyborg.shared.enums import SessionType, ProcessType, ProcessVersion, AppProtocol
from csle_cyborg.shared.observation import Observation
from csle_cyborg.simulator.state import State


# msf module is auxiliary/scanner/http/tomcat_mgr_login - then set RHOSTS and RPORT
class TomcatCredentialScanner(MSFScanner):
    def __init__(self, ip_address: IPv4Address, port: int, session: int, agent: str):
        super().__init__(session, agent)
        self.target = ip_address
        self.target_port = port

    def sim_execute(self, state: State):
        obs = Observation()
        obs.set_success(False)
        if self.session not in state.sessions[self.agent]:
            return obs
        from_host = state.sessions['Red'][self.session].host
        session = state.sessions['Red'][self.session]

        good = False
        if session.session_type == SessionType.MSF_SERVER and session.active:
            good = True

        if not good:
            obs.set_success(False)
            return obs

        session, from_interface = self.get_local_source_interface(local_session=session, remote_address=self.target)

        if from_interface is None:
            obs.set_success(False)
            return obs

        if str(self.target) == "127.0.0.1":
            target_host = from_host
        else:
            target_host = state.hosts[state.ip_addresses[self.target]]

        target_proc = None
        for process in target_host.processes:
            for conn in process.connections:
                if self.target_port == conn['local_port']:
                    if conn['local_address'] == IPv4Address("0.0.0.0") or (
                            conn['local_address'] == IPv4Address("127.0.0.1") and self.target == IPv4Address(
                            "127.0.0.1")) or conn['local_address'] == self.target:
                        target_proc = process
                    break

        if target_proc is None:
            obs.set_success(False)
            return obs
        else:
            if target_proc.process_type == ProcessType.WEBSERVER:
                if target_proc.version == ProcessVersion.APACHE_TOMCAT:
                    for conn in target_proc.connections:
                        if conn["local_port"] == self.target_port:
                            if conn['Application Protocol'] == AppProtocol.HTTP:
                                obs.set_success(True)
                                obs.add_process(hostid="hostid0", local_address=self.target,
                                                local_port=self.target_port,
                                                app_protocol="TCP", process_type=target_proc.process_type,
                                                process_version=target_proc.version, status="open")
                                obs.add_user_info(hostid="hostid0", username="tomcat", password="tomcat",
                                                  group_name="web server users")
                            else:
                                obs.set_success(False)
                                obs.add_process(hostid="hostid0", local_address=self.target,
                                                local_port=self.target_port,
                                                app_protocol="TCP", process_type=target_proc.process_type,
                                                status="open")
                else:
                    obs.set_success(False)
                    obs.add_process(hostid="hostid0", local_address=self.target, local_port=self.target_port,
                                    app_protocol="TCP", process_type=target_proc.process_type, status="open")
            else:
                obs.set_success(False)
        return obs

    def emu_execute(self, session_handler) -> Observation:
        obs = Observation()
        from csle_cyborg.CybORG import MSFSessionHandler
        if type(session_handler) is not MSFSessionHandler:
            obs.set_success(False)
            return obs
        output = session_handler.execute_module(mtype='auxiliary', mname='scanner/http/tomcat_mgr_login',
                                                opts={'RHOSTS': str(self.target),
                                                      'RPORT': self.target_port})
        obs.add_raw_obs(output)
        obs.set_success(False)
        return obs



# Copyright DST Group. Licensed under the MIT license.
import sys
from copy import deepcopy
from csle_cyborg.shared.action_space import ActionSpace
from csle_cyborg.shared.actions.action import Action
from csle_cyborg.shared.baseline_reward_calculator import BaselineRewardCalculator
from csle_cyborg.shared.blue_reward_calculator import HybridAvailabilityConfidentialityRewardCalculator
from csle_cyborg.shared.observation import Observation
from csle_cyborg.shared.red_reward_calculator import (DistruptRewardCalculator, PwnRewardCalculator,
                                                      HybridImpactPwnRewardCalculator)
from csle_cyborg.shared.results import Results
from csle_cyborg.shared.reward_calculator import RewardCalculator
from csle_cyborg.shared.reward_calculator import EmptyRewardCalculator

MAX_HOSTS = 5
MAX_PROCESSES = 100  # 50
MAX_CONNECTIONS = 10
MAX_VULNERABILITIES = 1
MAX_INTERFACES = 4
MAX_FILES = 10
MAX_SESSIONS = 10  # 80
MAX_USERS = 10
MAX_GROUPS = 10
MAX_PATCHES = 10


class AgentInterface:

    def __init__(self,
                 agent_class,
                 agent_name,
                 actions,
                 reward_calculator_type,
                 allowed_subnets,
                 scenario,
                 wrappers=None):
        self.hostname = {}
        self.username = {}
        self.group_name = {}
        self.process_name = {}
        self.interface_name = {}
        self.path = {}
        self.password = {}
        self.password_hash = {}
        self.file = {}
        self.actions = actions
        self.reward_calculator_type = reward_calculator_type
        self.last_action = None
        self.scenario = scenario
        self.reward_calculator = self.create_reward_calculator(
            self.reward_calculator_type, agent_name, scenario
        )
        self.agent_name = agent_name
        self.action_space = ActionSpace(self.actions, agent_name, allowed_subnets)
        self.agent = agent_class()
        if wrappers is not None:
            for wrapper in wrappers:
                if wrapper != 'None':
                    self.agent = getattr(sys.modules['CybORG.Agents.Wrappers'], wrapper)(agent=self.agent)
        self.agent.set_initial_values(
            action_space=self.action_space.get_max_action_space(),
            observation=Observation().data
        )

    def copy(self):
        self.action_space = self.action_space.copy()
        return deepcopy(self)

    def update(self, obs: dict, known=True):
        if isinstance(obs, Observation):
            obs = obs.data
        self.action_space.update(obs, known)

    def set_init_obs(self, init_obs, true_obs):
        if isinstance(init_obs, Observation):
            init_obs = init_obs.data
        if isinstance(true_obs, Observation):
            true_obs = true_obs.data
        self.update(true_obs, False)
        self.update(init_obs, True)
        self.reward_calculator.previous_state = true_obs
        self.reward_calculator.init_state = true_obs

        self.reward_calculator.previous_obs = init_obs
        self.reward_calculator.init_obs = init_obs

    def get_action(self, observation: Observation, action_space: dict = None):
        """Gets an action from the agent to perform on the environment"""
        if isinstance(observation, Observation):
            observation = observation.data
        if action_space is None:
            action_space = self.action_space.get_action_space()
        self.last_action = self.agent.get_action(observation, action_space)
        return self.last_action

    def train(self, result: Results):
        """Trains an agent with the new tuple from the environment"""
        if isinstance(result.observation, Observation):
            result.observation = result.observation.data
        if isinstance(result.next_observation, Observation):
            result.next_observation = result.next_observation.data
        result.action = self.last_action
        self.agent.train(result)

    def end_episode(self):
        self.agent.end_episode()
        self.reset()

    def reset(self):
        self.hostname = {}
        self.username = {}
        self.group_name = {}
        self.process_name = {}
        self.interface_name = {}
        self.path = {}
        self.password = {}
        self.password_hash = {}
        self.file = {}
        self.reward_calculator.reset()
        self.action_space.reset(self.agent_name)
        self.agent.end_episode()

    def create_reward_calculator(self, reward_calculator: str, agent_name: str, scenario) -> RewardCalculator:
        calc = None
        if reward_calculator == "Baseline":
            calc = BaselineRewardCalculator(agent_name)
        elif reward_calculator == 'Pwn':
            calc = PwnRewardCalculator(agent_name, scenario)
        elif reward_calculator == 'Disrupt':
            calc = DistruptRewardCalculator(agent_name, scenario)
        elif reward_calculator == 'None' or reward_calculator is None:
            calc = EmptyRewardCalculator(agent_name)
        elif reward_calculator == 'HybridAvailabilityConfidentiality':
            calc = HybridAvailabilityConfidentialityRewardCalculator(agent_name, scenario)
        elif reward_calculator == 'HybridImpactPwn':
            calc = HybridImpactPwnRewardCalculator(agent_name, scenario)
        else:
            raise ValueError(f"Invalid calculator selection: {reward_calculator} for agent {agent_name}")
        return calc

    def determine_reward(self, agent_obs: dict, true_obs: dict, action: Action, done: bool) -> float:
        return self.reward_calculator.calculate_reward(current_state=true_obs, action=action,
                                                       agent_observations=agent_obs, done=done)

    def get_observation_space(self):
        # returns the maximum observation space for the agent given its action set and the amount of parameters in the environment
        raise NotImplementedError



from csle_cyborg.shared.results import Results

class BaseAgent:

    def train(self, results: Results):
        """allows an agent to learn a policy"""
        raise NotImplementedError

    def get_action(self, observation, action_space):
        """gets an action from the agent that should be performed based on the agent's internal state and provided observation and action space"""
        raise NotImplementedError

    def end_episode(self):
        """Allows an agent to update its internal state"""
        raise NotImplementedError

    def set_initial_values(self, action_space, observation):
        raise NotImplementedError



# Copyright DST Group. Licensed under the MIT license.
import time
from csle_cyborg.shared.observation import Observation
from .shell_action import ShellAction


class ShellSleep(ShellAction):

    def __init__(self,
                 session: int = None,
                 agent: str = None,
                 t: int = 1):
        super().__init__(session, agent)
        self.t = t

    def sim_execute(self, state):
        return Observation()

    def emu_execute(self, session_handler):
        time.sleep(self.t)
        obs = Observation()
        obs.add_raw_obs(f"I slept {self.t} secs")
        return obs



# Copyright DST Group. Licensed under the MIT license.
from ipaddress import IPv4Address
from csle_cyborg.shared.actions.msf_actions_folder.remote_code_execution_folder.remote_code_execution import RemoteCodeExecution

# use msf module exploit/windows/smb/ms17_010_eternal_blue, set RHOSTS to target
# could also change LHOST, LPORT and RPORT (default 139)
# gives root session
from csle_cyborg.simulator.session import SessionType
from csle_cyborg.shared.enums import OperatingSystemType, ProcessType, ProcessVersion, OperatingSystemPatch
from csle_cyborg.shared.observation import Observation


class MSFEternalBlue(RemoteCodeExecution):
    def __init__(self, ip_address: IPv4Address, session: int, agent: str):
        super().__init__(session=session, agent=agent)
        self.target = ip_address

    def sim_execute(self, state):
        obs = Observation()
        obs.set_success(False)
        if self.session not in state.sessions[self.agent]:
            return obs
        session = state.sessions[self.agent][self.session]

        if session.session_type != SessionType.MSF_SERVER or not session.active:
            return obs

        # find shared subnet of the two hosts
        server_session, server_interface = self.get_local_source_interface(local_session=session, remote_address=self.target)

        if server_interface is None:
            return obs

        server_address = server_interface.ip_address

        if self.target == IPv4Address("127.0.0.1"):
            target_host = server_session.host
        else:
            target_host = state.hosts[state.ip_addresses[self.target]]

        # obs.add_interface_info(hostid='0', ip_address=server_address)
        # obs.add_interface_info(hostid='1', ip_address=self.target)

        # find out if smb is open
        smb_proc = None
        for proc in target_host.processes:
            if proc.process_type == ProcessType.SMB:
                # TODO: In case of SMB that is not the right version, should SMB process be in the obs?
                smb_proc = proc
                break

        # find out if smb is vulnerable (Windows OS + smb version)
        # Note that this exploit should actually work for all versions in the range Samba 3.0.20 - 3.0.25rc3
        if smb_proc is not None and smb_proc.version == ProcessVersion.SMBv1:
            obs.add_process(hostid="1", local_address=self.target, local_port=139, status="open",
                            process_type="smb")
            if target_host.os_type == OperatingSystemType.WINDOWS and OperatingSystemPatch.MS17_010:
                obs.set_success(True)

                root_user = None
                for u in state.hosts[state.ip_addresses[self.target]].users:
                    if u.username == "SYSTEM":
                        root_user = u

                new_session = state.add_session(host=target_host.hostname, agent=self.agent,
                                                user=root_user, session_type="meterpreter", parent=server_session)

                local_port = target_host.get_ephemeral_port()
                new_connection = {"local_port": local_port,
                                  "Application Protocol": "tcp",
                                  "remote_address": server_address,
                                  "remote_port": 4444,
                                  "local_address": self.target
                                  }
                new_session.process.connections.append(new_connection)

                target_host.add_process(name="telnet", ppid=1, path="/usr/bin/", user=root_user)
                remote_port = {"local_port": 4444,
                               "Application Protocol": "tcp",
                               "local_address": server_address,
                               "remote_address": self.target,
                               "remote_port": local_port
                               }

                server_session.process.connections.append(remote_port)

                obs.add_process(hostid="0", local_address=server_address, remote_address=str(self.target),
                                local_port=4444, remote_port=local_port)
                obs.add_process(hostid="1", local_address=str(self.target), remote_address=server_address,
                                local_port=local_port, remote_port=4444)
                obs.add_session_info(hostid="1", session_id=new_session.ident, session_type=new_session.session_type, agent=self.agent)
            else:
                obs.add_interface_info(ip_address=str(self.target))

        return obs

    def emu_execute(self, session_handler) -> Observation:
        obs = Observation()
        from csle_cyborg.CybORG import MSFSessionHandler
        if type(session_handler) is not MSFSessionHandler:
            obs.set_success(False)
            return obs
        output = session_handler.execute_module(mtype='exploit', mname='windows/smb/ms17_010_eternalblue', opts={'RHOSTS': str(self.target)}, payload_name='generic/shell_bind_tcp')
        obs.add_raw_obs(output)
        obs.set_success(False)
        session_handler._log_debug(output)
        """ Example:
        [*] Started reverse TCP handler on 10.0.5.31:4444 
        [*] 10.0.14.27:445 - Connecting to target for exploitation.
        [+] 10.0.14.27:445 - Connection established for exploitation.
        [+] 10.0.14.27:445 - Target OS selected valid for OS indicated by SMB reply
        [*] 10.0.14.27:445 - CORE raw buffer dump (51 bytes)
        [*] 10.0.14.27:445 - 0x00000000  57 69 6e 64 6f 77 73 20 53 65 72 76 65 72 20 32  Windows Server 2
        [*] 10.0.14.27:445 - 0x00000010  30 30 38 20 52 32 20 53 74 61 6e 64 61 72 64 20  008 R2 Standard 
        [*] 10.0.14.27:445 - 0x00000020  37 36 30 31 20 53 65 72 76 69 63 65 20 50 61 63  7601 Service Pac
        [*] 10.0.14.27:445 - 0x00000030  6b 20 31                                         k 1             
        [+] 10.0.14.27:445 - Target arch selected valid for arch indicated by DCE/RPC reply
        [*] 10.0.14.27:445 - Trying exploit with 12 Groom Allocations.
        [*] 10.0.14.27:445 - Sending all but last fragment of exploit packet
        [*] 10.0.14.27:445 - Starting non-paged pool grooming
        [+] 10.0.14.27:445 - Sending SMBv2 buffers
        [+] 10.0.14.27:445 - Closing SMBv1 connection creating free hole adjacent to SMBv2 buffer.
        [*] 10.0.14.27:445 - Sending final SMBv2 buffers.
        [*] 10.0.14.27:445 - Sending last fragment of exploit packet!
        [*] 10.0.14.27:445 - Receiving response from exploit packet
        [+] 10.0.14.27:445 - ETERNALBLUE overwrite completed successfully (0xC000000D)!
        [*] 10.0.14.27:445 - Sending egg to corrupted connection.
        [*] 10.0.14.27:445 - Triggering free of corrupted buffer.
        [*] Command shell session 3 opened (10.0.5.31:4444 -> 10.0.14.27:49400) at 2020-08-07 06:09:46 +0000
        [+] 10.0.14.27:445 - =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
        [+] 10.0.14.27:445 - =-=-=-=-=-=-=-=-=-=-=-=-=-WIN-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
        [+] 10.0.14.27:445 - =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
        """
        for line in output.split(('\n')):
            if '[*] Command shell session' in line:
                obs.set_success(True)
                #print(list(enumerate(line.split(' '))))
                split = line.split(' ')
                session = split[4]
                rip, rport = split[6].replace('(', '').split(':')
                lip, lport = split[8].replace(')', '').split(':')
                obs.add_session_info(hostid=str(self.target), session_id=session, session_type='msf_shell')
                obs.add_process(hostid=str(self.target), local_address=lip, local_port=lport, remote_address=rip, remote_port=rport)
        return obs



import inspect

from stable_baselines3 import PPO

from csle_cyborg.agents.simple_agents.base_agent import BaseAgent
from csle_cyborg.agents.wrappers.challenge_wrapper import ChallengeWrapper

class BlueLoadAgent(BaseAgent):
    # agent that loads a StableBaselines3 PPO model file
    def train(self, results):
        pass

    def end_episode(self):
        pass

    def set_initial_values(self, action_space, observation):
        pass

    def __init__(self, model_file: str = None):
        if model_file is not None:
            self.model = PPO.load(model_file)
        else:
            self.model = None

    def get_action(self, observation, action_space):
        """gets an action from the agent that should be performed based on the agent's internal state and provided observation and action space"""
        if self.model is None:
            from csle_cyborg.main import Main
            path = str(inspect.getfile(Main))
            path = path[:-7] + '/Shared/Scenarios/Scenario1b.yaml'
            cyborg = ChallengeWrapper(env=Main(path, 'sim'), agent_name='Blue')
            self.model = PPO('MlpPolicy', cyborg)
        action, _states = self.model.predict(observation)
        return action



from typing import List

from csle_cyborg.shared.actions.action import Action
from csle_cyborg.simulator.subnet import Subnet


class ConcreteAction(Action):
    """
    Abstract class for all actions that directly simulate real world commands/tools.

    Any action attempting to simulate a real world command or tool should inherit from this class or one of its
    children.
    """
    def __init__(self,session: int,agent: str):
        super().__init__()
        self.agent = agent
        self.session = session

    def check_routable(self, from_subnets: List[Subnet], to_subnets: List[Subnet]) -> dict:
        """
        Checks which ports in from_subnets can be accessed by hosts in to_subnets.

        Checks NACL data to see if any ports are blocked.
        """
        # check what ports from subnets allow to any to subnets
        ports = {} # port: (to_subnet, from_subnet)
        for from_subnet in from_subnets:
            for to_subnet in to_subnets:
                # check if traffic from subnet is stopped by to subnet nacl
                if from_subnet.name in to_subnet.nacls:
                    if 'ICMP' not in ports:
                       ports['ICMP'] = (from_subnet.cidr, to_subnet.cidr)
                    if 'all' in to_subnet.nacls[from_subnet.name]['in']:
                        # if all ports accepted in then set ports to all and we are done
                        return {'all': (from_subnet.cidr, to_subnet.cidr)}
                    elif 'None' in to_subnet.nacls[from_subnet.name]['in']:
                        # If you don't have access to Enteprise network, you can't act on Operational Host
                        # TODO refactor this hacky fix
                        permission = self.check_for_enterprise_sessions()
                        ports = {'all': (from_subnet.cidr, to_subnet.cidr)} if permission else {}
                        return ports
                        
                    else:
                        # we only add the ports in rules to our accepted ports
                        for rule in to_subnet.nacls[from_subnet.name]['in']:
                            if rule['PortRange'] is int and rule['PortRange'] not in ports:
                                ports[rule["PortRange"]] = (from_subnet.cidr, to_subnet.cidr)
                            else:
                                for p in range(rule["PortRange"][0], rule["PortRange"][1]):
                                    if p not in ports:
                                        ports[p] = (from_subnet.cidr, to_subnet.cidr)
                elif 'all' in to_subnet.nacls:
                    if 'ICMP' not in ports:
                        ports['ICMP'] = (from_subnet.cidr, to_subnet.cidr)
                    # if all ports accepted out then use inbound rules only
                    if 'all' in to_subnet.nacls['all']['in']:
                        # if all ports accepted in then set ports to all and we are done
                        return {'all': (from_subnet.cidr, to_subnet.cidr)}
                    else:
                        # we only add the ports in rules to our accepted ports
                        for rule in to_subnet.nacls['all']['in']:
                            if rule['PortRange'] is int and rule['PortRange'] not in ports:
                                ports[rule["PortRange"]] = (from_subnet.cidr, to_subnet.cidr)
                            else:
                                for p in range(rule["PortRange"][0], rule["PortRange"][1]):
                                    if p not in ports:
                                        ports[p] = (from_subnet.cidr, to_subnet.cidr)
                else:
                    # this means that traffic cannot reach move between these 2 subnets
                    continue

        return ports

    def check_for_enterprise_sessions(self):
        permission = False
        for session_id in self.state.sessions[self.agent]:
            session = self.state.sessions[self.agent][session_id]
            if 'Enterprise' in session.host:
                permission = True

        return permission





## The following code contains work of the United States Government and is not subject to domestic copyright protection under 17 USC § 105.
## Additionally, we waive copyright and related rights in the utilized code worldwide through the CC0 1.0 Universal public domain dedication.

import random
from ipaddress import IPv4Address
from abc import ABC, abstractmethod
from typing import Optional

from csle_cyborg.shared.observation import Observation
from csle_cyborg.shared.actions.action import Action
from csle_cyborg.shared.actions.concrete_actions.exploit_action import ExploitAction
from csle_cyborg.shared.actions.concrete_actions.blue_keep import BlueKeep
from csle_cyborg.shared.actions.concrete_actions.eternal_blue import EternalBlue
from csle_cyborg.shared.actions.concrete_actions.http_rfi import HTTPRFI
from csle_cyborg.shared.actions.concrete_actions.http_srfi import HTTPSRFI
from csle_cyborg.shared.actions.concrete_actions.ssh_brute_force import SSHBruteForce
from csle_cyborg.shared.actions.concrete_actions.ftp_directory_traversal import FTPDirectoryTraversal
from csle_cyborg.shared.actions.concrete_actions.sql_injection import SQLInjection
from csle_cyborg.shared.actions.concrete_actions.haraka_rce import HarakaRCE
from csle_cyborg.simulator.session import RedAbstractSession
from csle_cyborg.simulator.state import State

# pylint: disable=too-few-public-methods
class ExploitActionSelector(ABC):
    """
    Examines the target host and returns a selected applicable escalate action
    if any, as well as processes that are required to be genuine
    """
    # pylint: disable=missing-function-docstring
    @abstractmethod
    def get_exploit_action(self, *, state: State, session: int,
            agent: str, ip_address: IPv4Address) -> Optional[ExploitAction]:
        pass

class DefaultExploitActionSelector(ExploitActionSelector):
    """
    Chooses exploit based on available ports
    """
    def get_exploit_action(self, *, state: State, session: int,
            agent: str, ip_address: IPv4Address, priority=None) -> Optional[ExploitAction]:
        ports = state.sessions[agent][session].ports[ip_address]
        # Weights need to be different. Should be refactored to a list.
        exploit_options = {EternalBlue: 2.0 if 139 in ports else None,
                            BlueKeep: 1.0 if 3389 in ports else None,
                            HTTPRFI: 3.0 if 80 in ports else None,
                            HTTPSRFI: 4.0 if 443 in ports else None,
                            SSHBruteForce: 0.1 if 22 in ports else None,
                            SQLInjection: 5.0 if (3390 in ports) and (80 in ports or 443 in ports) \
                                    else None,
                            HarakaRCE: 6.0 if 25 in ports else None,
                            FTPDirectoryTraversal: 7.0 if 21 in ports else None,
                            }
        exploit_options = {a: x for a, x in exploit_options.items() if x is not None}
        if len(exploit_options) > 1:
            # use information to populate weights for which exploit to select
            odds_of_top_choice = 0.75
            top_choice = max(list(exploit_options.values()))
            other_choice = random.choices([x for x in list(exploit_options.values()) if x != top_choice])[0]
            # other_choice = random.choices([x for x in list(exploit_options.values())])[0]
            priority_value = exploit_options.get(priority)
            if priority_value is not None:
                selected_choice = priority_value
            elif random.random() < odds_of_top_choice:
                selected_choice = top_choice
            else:
                selected_choice = other_choice

            # sample the exploit to perform from the exploit weights
            sub_action = list(exploit_options.keys())[
                    list(exploit_options.values()).index(selected_choice)](session=session,
                                                                                            agent=agent,
                                                                                            ip_address=ip_address,
                                                                                            target_session=session)
            return sub_action
        elif len(exploit_options) == 1:
            return list(exploit_options.keys())[0](session=session,
                                                                                            agent=agent,
                                                                                            ip_address=ip_address,
                                                                                            target_session=session)
        else:
            return None


_default_exploit_action_selector = DefaultExploitActionSelector()

class ExploitRemoteService(Action):
    def __init__(self, session: int, agent: str, ip_address: IPv4Address, priority=None):
        super().__init__()
        self.ip_address = ip_address
        self.agent = agent
        self.session = session
        self.exploit_action_selector = _default_exploit_action_selector
        self.priority = priority

    def sim_execute(self, state: State) -> Observation:
        # find session inside or close to the target subnet
        if type(state.sessions[self.agent][self.session]) is not RedAbstractSession or self.ip_address not in state.sessions[self.agent][self.session].ports:
            return Observation(success=False)

        self.sub_action = self.exploit_action_selector.get_exploit_action(state=state,
                session=self.session, agent=self.agent, ip_address=self.ip_address, priority=self.priority)
        if self.sub_action is None:
            return Observation(success=False)
        obs = self.sub_action.sim_execute(state)
        if self.ip_address != IPv4Address('127.0.0.1') and obs.data['success'] == True:
            hostname = obs.data[str(self.ip_address)]["System info"]["Hostname"]
            os = obs.data[str(self.ip_address)]["System info"]["OSType"]
            state.sessions[self.agent][self.session].addos(hostname, os)
        return obs

    def __str__(self):
        return f"{self.__class__.__name__} {self.ip_address}"



# Copyright DST Group. Licensed under the MIT license.
from csle_cyborg.shared.actions.session_action import SessionAction
from csle_cyborg.shared.observation import Observation


class HostMonitoringAction(SessionAction):
    """Abstract class for a host monitoring action. """

    def __init__(self, session: int):
        """
        Parameters
        ----------
        session : int
            the id of the session to perform action in
        """
        super().__init__(session)
        self.name = self.__class__.__name__

    def emu_execute(self,
                    session_handler,
                    *args,
                    **kwargs):
        """Execute and action in emulator environment

        Parameters
        ----------
        session_handler : SessionHandler
           session handler object for game session of action (i.e. that
           matches session_id)

        Returns
        -------
        Observation
            Result of performing action
        """
        raise NotImplementedError

    def parse(self, results) -> Observation:
        """Parses the results of the execute action to create an observation"""
        raise NotImplementedError

    def sim_execute(self, state):
        raise NotImplementedError



# Copyright DST Group. Licensed under the MIT license.

from csle_cyborg.shared.observation import Observation
from csle_cyborg.shared.actions.global_actions.global_action import GlobalAction


class ShutdownGame(GlobalAction):
    """Shutdown a game. """

    def __init__(self, game_id: int, teardown: bool = True):
        """
        Parameters
        ----------
        game_id : int
            ID of game to shutdown
        teardown : bool, optional
            whether to terminate network resources or not (default=True)
        """
        super().__init__()
        self.game_id = game_id
        self.teardown = teardown

    def emu_execute(self, team_server, *args, **kwargs) -> Observation:
        obs = Observation()
        success = team_server.shutdown_game(self.game_id, self.teardown)
        obs.set_success(success)
        return obs



# Copyright DST Group. Licensed under the MIT license.
import datetime
from ipaddress import IPv4Address, IPv4Network
from csle_cyborg.shared import enums as CyEnums


class NetworkInterface:
    """A class for storing network interface information """

    def __init__(self,
                 hostid: str = None,
                 interface_name: str = None,
                 ip_address: IPv4Address = None,
                 subnet: IPv4Network = None):
        self.hostid = hostid
        self.interface_name = interface_name
        self.ip_address = IPv4Address(ip_address)
        self.subnet = IPv4Network(subnet)

    def get_info(self) -> dict:
        """Return network interface as dict.

        Keys of dict match arguments of Observation.add_interface_info()
        """
        return {
            "hostid": self.hostid,
            "interface_name": self.interface_name,
            "ip_address": self.ip_address,
            "subnet": self.subnet
        }

    def __str__(self):
        output = [f"{self.__class__.__name__}:"]
        for k, v in self.get_info().items():
            if v is None:
                continue
            output.append(f"{k}={v}")
        return f" ".join(output)


class File:
    """A dataclass for storing information about a single file """

    def __init__(self,
                 name: str,
                 path: str,
                 file_type: str = None,
                 vendor: str = None,
                 version: str = None,
                 user: str = None,
                 user_permissionss: int = None,
                 group: str = None,
                 group_permissions: int = None,
                 default_permissions: int = None,
                 last_modified_time: datetime.datetime = None):
        self.name = name
        self.path = path
        self.user = user
        self.group = group
        self.vendor = vendor

        self.user_permissionss = user_permissionss
        if self.user_permissionss is None and self.user is not None:
            self.group_permissions = 7

        self.group_permissions = group_permissions
        if self.group_permissions is None and self.group is not None:
            self.group_permissions = 7

        self.default_permissions = default_permissions
        if self.default_permissions is None:
            self.default_permissions = 7

        self.last_modified_time = last_modified_time
        if self.last_modified_time is not None:
            self.last_modified_time = datetime.strptime(
                self.last_modified_time, "%d %b %Y %H:%M"
            )

        self.file_type = file_type
        if file_type is None:
            self.file_type = CyEnums.FileType.UNKNOWN
        elif not isinstance(file_type, CyEnums.FileType):
            self.file_type = CyEnums.FileType.parse_string(file_type)

        self.version = version
        if version is not None and not isinstance(CyEnums.FileVersion):
            self.version = CyEnums.FileVersion.parse_string(version)

    def get_info(self) -> dict:
        """Return network interface as dict.

        Keys of dict match arguments of Observation.add_file_info()
        """
        return {
            "path": self.path,
            "name": self.name,
            "vendor": self.vendor,
            "version": self.version,
            "file_type": self.file_type,
            "user": self.user,
            "user_permissions": self.user_permissionss,
            "group": self.group,
            "group_permissions": self.group_permissions,
            "default_permissions": self.default_permissions,
            "last_modified_time": self.last_modified_time
        }

    def __str__(self):
        output = [f"{self.__class__.__name__}:"]
        for k, v in self.get_info().items():
            if v is None:
                continue
            output.append(f"{k}={v}")
        return f" ".join(output)


class Credentials:
    """A class for storing a set of credentials """

    def __init__(self,
                 username: str,
                 password: str = None,
                 key_path: str = None,
                 password_hash: str = None,
                 password_hash_type: str = None,
                 groups: list = None):
        self.username = username
        self.password = password
        self.key_path = key_path
        self.password_hash = password_hash
        self.password_hash_type = password_hash_type
        self.groups = [] if groups is None else groups

    def get_info(self) -> dict:
        """Return credentials as dict

        Keys of dict match arguments of Observation.add_user_info()
        """
        return {
            "username": self.username,
            "password": self.password,
            "password_hash": self.password_hash,
            "password_hash_type": self.password_hash_type,
            "key_path": self.key_path
        }

    def __str__(self):
        output = [f"{self.__class__.__name__}:"]
        for k, v in self.get_info().items():
            if v is None:
                continue
            output.append(f"{k}={v}")
        return f" ".join(output)


class OperatingSystemInfo:
    """A class for storing information about the OS of a VM """

    def __init__(self,
                 os_type: CyEnums.OperatingSystemType = None,
                 dist: CyEnums.OperatingSystemDistribution = None,
                 version: CyEnums.OperatingSystemVersion = None,
                 kernel: CyEnums.OperatingSystemKernelVersion = None,
                 architecture: CyEnums.Architecture = None,
                 patch: CyEnums.OperatingSystemPatch = None):
        self.os_type = os_type
        self.dist = dist
        self.version = version
        self.kernel = kernel
        self.architecture = architecture
        self.patch = patch

    def get_info(self) -> dict:
        """Return OS info as dict

        Keys of dict match arguments of Observation.add_system_info()
        """
        return {
            "os_type": self.os_type,
            "os_distribution": self.dist,
            "os_verson": self.version,
            "os_kernel": self.kernel,
            "os_patches": self.patch,
            "architecture": self.architecture
        }

    def __str__(self):
        output = [f"{self.__class__.__name__}:"]
        for k, v in self.get_info().items():
            if v is None:
                continue
            output.append(f"{k}={v}")
        return f" ".join(output)


class Image:
    """An class for storing VM Image information """

    def __init__(self,
                 name: str,
                 services: list = None,
                 os_info: OperatingSystemInfo = None,
                 credentials: dict = None,
                 root_user: str = None,
                 key_access: bool = False,
                 files: dict = None,
                 aux_info: dict = None):
        """
        Parameters
        ----------
        name : str
            The name of the image. This is used to distinguish between images
            with the same OS type, distribution and version in a human
            readable format. e.g. between standard ubuntu 14.04 and the
            Metasploitable 3 ubuntu 14.04
        services : list, optional
            Service objects defining services running on machine (default=None)
        os_info : str, optional
            image os information (i.e. type, distribution, version)
            (default=None)
        credentials : dict, optional
            map of user to credentials for the VM image (default=None)
        root_user : str, optional
            the root user for the image. This is the user whose credentials are
            used when configuring any instances using this image (default=None)
        key_access : bool, optional
            whether SSH access to instance is restricted to key only
            (default=False)
        files : dict, optional
            any known/specified files that are on the image (default=None)
        aux_info: dict, optional
            any extra Image specific information (e.g. MSF or Host monitoring
            info0 (Default=None)
        """
        self.name = name
        self.services = services
        self.os_info = OperatingSystemInfo if os_info is None else os_info
        self.credentials = {} if credentials is None else credentials
        self.root_user = root_user
        self.key_access = key_access
        self.files = {} if files is None else files
        self.aux_info = {} if aux_info is None else aux_info
        assert root_user is None or root_user in credentials, \
            f"Root user of Image must have matching entry in credentials"

    def get_root_user_creds(self) -> Credentials:
        """Get the credentials of the root user of Image.

        Returns
        -------
        Credentials
            Root user credentials

        Raises
        ------
        AttributeError
            If no valid credentials can be found

        Notes
        -----
        If root_user attribute of image is not defined, this will return the
        first user in the image credentials dict
        """
        if self.root_user is not None:
            return self.credentials[self.root_user]
        for username, creds in self.credentials.items():
            if not self.key_access or creds.key_path is not None:
                return creds
        raise AttributeError("No valid root user credentials found for {self}")

    def __str__(self):
        creds = [f"{u}: {c}" for u, c in self.credentials.items()]
        creds_str = "[" + ", ".join(creds) + "]"
        output = [f"{self.__class__.__name__}:",
                  f"Name={self.name}"
                  f"Services={self.services}"
                  f"OS Info={str(self.os_info)}"
                  f"Credentials={creds_str}",
                  f"Root User={self.root_user}"
                  f"Key Access={self.key_access}",
                  f"Files={self.files}",
                  f"Aux Info={self.aux_info}"]
        return "  ".join(output)

    def __eq__(self, other):
        if not isinstance(other, Image):
            return False
        return (other.name == self.name
                and other.services == self.services
                and other.os_info == self.os_info
                and other.credentials == self.credentials
                and other.files == self.files
                and other.aux_info == self.aux_info)



# Copyright DST Group. Licensed under the MIT license.

from csle_cyborg.shared.observation import Observation
from csle_cyborg.simulator.session import SessionType
from csle_cyborg.shared.enums import QueryType
from .velociraptor_action import VelociraptorAction


class VelociraptorPoll(VelociraptorAction):

    def __init__(self, session: int, agent: str):
        super().__init__(session=session,
                         agent=agent,
                         query_type=QueryType.SYNC,
                         poll_alerts=True)

    def sim_execute(self, state):
        obs = Observation()
        obs.set_success(False)
        if self.session not in state.sessions[self.agent]:
            return obs
        session = state.sessions[self.agent][self.session]

        if session.session_type != SessionType.VELOCIRAPTOR_SERVER:
            return obs

        obs = state.get_true_state()
        obs.set_success(True)
        # remove hosts without a velociraptor client or server from obs
        for hostname, host in state.hosts.items():
            client = False
            # need to check if key self.agent exists in dict.sessions first
            if self.agent in host.sessions:
                for session in host.sessions[self.agent]:
                    s_type = session.session_type
                    if s_type == SessionType.VELOCIRAPTOR_CLIENT \
                       or s_type == SessionType.VELOCIRAPTOR_SERVER:
                        client = True
                        break
            if not client:
                obs.data.pop(hostname)

        # remember to remove red sessions from white obs
        for host, hostinfo in obs.get_dict().items():
            if host != "success":
                for session in hostinfo["Sessions"]:
                    if session["Agent"] != self.agent:
                        hostinfo["Sessions"].remove(session)
        return obs

    def emu_execute(self,
                    session_handler,
                    *args,
                    **kwargs) -> Observation:
        vel_controller = session_handler.controller
        return vel_controller.get_latest_alerts()



from csle_cyborg.shared.observation import Observation
from csle_cyborg.shared.actions.concrete_actions.concrete_action import ConcreteAction
from csle_cyborg.simulator.host import Host
from csle_cyborg.simulator.process import Process
from csle_cyborg.simulator.state import State


class StopProcess(ConcreteAction):
    def __init__(self, session: int, agent: str, target_session: int, pid: int):
        super(StopProcess, self).__init__(session, agent)
        self.pid = pid
        self.target_session = target_session

    def sim_execute(self, state: State) -> Observation:
        obs = Observation()
        if self.session not in state.sessions[self.agent] or self.target_session not in state.sessions[self.agent]:
            obs.set_success(False)
            return obs
        target_host: Host = state.hosts[state.sessions[self.agent][self.target_session].host]
        session = state.sessions[self.agent][self.session]
        target_session = state.sessions[self.agent][self.target_session]

        if not session.active or not target_session.active:
            obs.set_success(False)
            return obs
        proc = target_host.get_process(self.pid)
        if proc is not None:
            if proc.user != 'root' and proc.user != 'SYSTEM':
                obs.set_success(True)
                self.kill_process(state, target_host, proc)
            else:
                obs.set_success(False)
        else:
            obs.set_success(False)
        return obs

    def kill_process(self, state: State, host: Host, process: Process):
        agent, session = state.get_session_from_pid(host.hostname, pid=process.pid)
        host.processes.remove(process)
        if process.pid in [i['process'] for i in host.services.values()]:
            process.pid = None
            host.add_process(**process.__dict__)
            service = True
        else:
            service = False
        if session is not None:
            host.sessions[agent].remove(session)
            state.sessions[agent].pop(session)
            if service:
                session_reloaded = state.add_session(host=host.hostname, user=session.user,
                                                    session_type=session.session_type, agent=session.agent,
                                                    parent=session.parent, timeout=session.timeout)



## The following code contains work of the United States Government and is not subject to domestic copyright protection under 17 USC § 105.
## Additionally, we waive copyright and related rights in the utilized code worldwide through the CC0 1.0 Universal public domain dedication.
from ipaddress import IPv4Address
from typing import List
from csle_cyborg.shared.enums import ProcessType, ProcessVersion, TransportProtocol, DecoyType
from csle_cyborg.simulator.entity import Entity


class Process(Entity):
    def __init__(self, process_name: str, pid: int, parent_pid: int, username: str, program_name: str = None,
                 path: str = None, open_ports: list = None, process_type: str = None, process_version: str = None,
                 decoy_type: DecoyType = DecoyType.NONE, properties: List[str] = None):
        """
        :param process_name: name of process
        :param pid: id of process
        :param parent_pid: id of parent of process
        :param program_name: program the process is running
        :param username: the user runnning the process
        :param path: the path of the program the process is running
        :param open_ports: listening ports of structure [{Port: int, Address: str, Application Protocol: str}, ...]
        :param process_type: the type of process
        :param process_version: the version of the process
        :param decoy_type: which red actions are prevented despite appearing vulnerable
        :param properties: properties of the process to specify configuration details such as RFI presence
        """
        super().__init__()
        self.name = process_name
        self.pid = pid
        self.ppid = parent_pid
        self.program = program_name
        self.user = username
        self.path = path
        self.open_ports = open_ports
        self.decoy_type = decoy_type
        self.connections = []  # Connections has the structure [{local_port, local_address, remote_port, Remote Address, Application Protocol}]
        if properties is None:
            self.properties = []
        else:
            self.properties = properties
        if open_ports is not None:
            for port_dict in open_ports:
                interface = port_dict['local_address']
                if interface == 'broadcast':
                    interface = IPv4Address('0.0.0.0')
                elif interface == 'local':
                    interface = IPv4Address('127.0.0.1')
                else:
                    interface = IPv4Address(interface)
                transport_protocol = port_dict.get("transport_protocol", 'UNKNOWN')
                if type(transport_protocol) is not TransportProtocol:
                    transport_protocol = TransportProtocol.parse_string(transport_protocol)
                new_connection = {'local_port': port_dict['local_port'], 'local_address': interface,
                                  'transport_protocol': transport_protocol}

                self.connections.append(new_connection)

        self.process_type = None
        if process_type is not None:
            if type(process_type) is str:
                self.process_type = ProcessType.parse_string(process_type)
            else:
                self.process_type = process_type
        elif process_name is not None:
            self.process_type = ProcessType.parse_string(process_name)

        if process_version is not None:
            self.version = ProcessVersion.parse_string(process_version)
        else:
            self.version = None

    def get_state(self):
        observations = []
        for connections_dict in self.connections:
            obs = {"pid": self.pid, "parent_pid": self.ppid, "process_name": self.name, "program_name": self.program,
                   "path": self.path, "process_type": self.process_type,
                   "process_version": self.version, "local_port": connections_dict['local_port'],
                   "local_address": connections_dict["local_address"]}
            if "remote_port" in connections_dict:
                obs["remote_port"] = connections_dict["remote_port"]
            if "remote_address" in connections_dict:
                obs["remote_address"] = connections_dict["remote_address"]
            if "transport_protocol" in connections_dict:
                obs["transport_protocol"] = connections_dict["transport_protocol"]
            if self.user is not None:
                obs["username"] = self.user
            observations.append(obs)
        if not observations:
            obs = {"pid": self.pid,
                   "parent_pid": self.ppid,
                   "process_name": self.name,
                   "program_name": self.program,
                   "path": self.path,
                   "process_type": self.process_type,
                   "process_version": self.version,
                   'username': self.user}
            observations.append(obs)
        return observations

    def __str__(self):
        return f'{self.name}: {self.pid} <- {self.ppid}'



# Copyright DST Group. Licensed under the MIT license.
from csle_cyborg.shared.observation import Observation
from csle_cyborg.shared.actions.global_actions.global_action import GlobalAction


class GlobalEcho(GlobalAction):

    def __init__(self, echo_cmd):
        super().__init__()
        self.cmd = echo_cmd

    def emu_execute(self, team_server) -> Observation:
        raise ValueError("Not implemented")



import subprocess
import inspect
import time
from statistics import mean, stdev
from csle_cyborg.main import Main
from csle_cyborg import CYBORG_VERSION
from csle_cyborg.agents.simple_agents.b_line import B_lineAgent
from csle_cyborg.agents.simple_agents.sleep_agent import SleepAgent
from csle_cyborg.agents.simple_agents.blue_load_agent import BlueLoadAgent
from csle_cyborg.agents.simple_agents.meander import RedMeanderAgent
from csle_cyborg.agents.wrappers.challenge_wrapper import ChallengeWrapper

MAX_EPS = 100
agent_name = 'Blue'


def wrap(env):
    return ChallengeWrapper(env=env, agent_name='Blue')

def get_git_revision_hash() -> str:
    return subprocess.check_output(['git', 'rev-parse', 'HEAD']).decode('ascii').strip()

if __name__ == "__main__":
    cyborg_version = CYBORG_VERSION
    scenario = 'Scenario2'
    commit_hash = get_git_revision_hash()
    # ask for a name
    name = input('Name: ')
    # ask for a team
    team = input("Team: ")
    # ask for a name for the agent
    name_of_agent = input("Name of technique: ")

    lines = inspect.getsource(wrap)
    wrap_line = lines.split('\n')[1].split('return ')[1]

    # Change this line to load your agent
    agent = BlueLoadAgent()

    print(f'Using agent {agent.__class__.__name__}, if this is incorrect please update the code to load in your agent')

    file_name = str(inspect.getfile(Main))[:-10] + '/Evaluation/' + time.strftime("%Y%m%d_%H%M%S") + f'_{agent.__class__.__name__}.txt'
    print(f'Saving evaluation results to {file_name}')
    with open(file_name, 'a+') as data:
        data.write(f'CybORG v{cyborg_version}, {scenario}, Commit Hash: {commit_hash}\n')
        data.write(f'author: {name}, team: {team}, technique: {name_of_agent}\n')
        data.write(f"wrappers: {wrap_line}\n")

    path = str(inspect.getfile(Main))
    path = path[:-7] + f'/Shared/Scenarios/{scenario}.yaml'

    print(f'using CybORG v{cyborg_version}, {scenario}\n')
    for num_steps in [30, 50, 100]:
        for red_agent in [B_lineAgent, RedMeanderAgent, SleepAgent]:

            cyborg = Main(path, 'sim', agents={'Red': red_agent})
            wrapped_cyborg = wrap(cyborg)

            observation = wrapped_cyborg.reset()
            # observation = cyborg.reset().observation

            action_space = wrapped_cyborg.get_action_space(agent_name)
            # action_space = cyborg.get_action_space(agent_name)
            total_reward = []
            actions = []
            for i in range(MAX_EPS):
                r = []
                a = []
                # cyborg.env.env.tracker.render()
                for j in range(num_steps):
                    action = agent.get_action(observation, action_space)
                    observation, rew, done, info = wrapped_cyborg.step(action)
                    # result = cyborg.step(agent_name, action)
                    r.append(rew)
                    # r.append(result.reward)
                    a.append((str(cyborg.get_last_action('Blue')), str(cyborg.get_last_action('Red'))))
                agent.end_episode()
                total_reward.append(sum(r))
                actions.append(a)
                # observation = cyborg.reset().observation
                observation = wrapped_cyborg.reset()
            print(f'Average reward for red agent {red_agent.__name__} and steps {num_steps} is: {mean(total_reward)} with a standard deviation of {stdev(total_reward)}')
            with open(file_name, 'a+') as data:
                data.write(f'steps: {num_steps}, adversary: {red_agent.__name__}, mean: {mean(total_reward)}, standard deviation {stdev(total_reward)}\n')
                for act, sum_rew in zip(actions, total_reward):
                    data.write(f'actions: {act}, total reward: {sum_rew}\n')



# Copyright DST Group. Licensed under the MIT license.

from csle_cyborg.shared.observation import Observation
from csle_cyborg.simulator.session import SessionType
from .velociraptor_artifact_action import VelociraptorArtifactAction


class GetProcessList(VelociraptorArtifactAction):
    """Get a list of all the process IDs for processes running on the host

    Velociraptor Reference
    ----------------------
    Artifact: https://www.velocidex.com/docs/artifacts/linux/#linuxsyspslist
    """

    def __init__(self, session: int, agent: str, hostname: str):
        super().__init__(session=session,
                         hostname=hostname,
                         agent=agent,
                         artifact='Linux.Sys.Pslist',
                         env={},
                         flow_completion_wait_limit=90)

    def sim_execute(self, state):
        obs = Observation()
        obs.set_success(False)
        if self.session not in state.sessions[self.agent]:
            return obs
        session = state.sessions[self.agent][self.session]

        if session.session_type != SessionType.VELOCIRAPTOR_CLIENT:
            return obs

        obs.set_success(True)

        for process in session.host.processes:
            obs.add_process(
                hostid="0",
                pid=process.pid,
                process_name=process.name
            )

        return obs

    def parse(self, results: list, client_id: str) -> Observation:
        """Parses the results of the execute action to create an observation

        Each result should be in the form:

        {
            'Pid': int,
            'Ppid': int,
            'Name': str,
            'Cmdline': str,
            'Exe': str,
            'Hash': None or dict,
            'Username': str,
            'CreatedTime': timestamp str,
            'RSS': int,
            'Deleted': bool
        }

        Hash dict in form:
        {
            hash_type: hash
        }
        Where hash_type and hash are both strings:

        e.g.
        'Hash': {
            'MD5': 'eb5c...',
            'SHA1': 'dbaad...',
            'SHA256': '6045...''
        }

        """
        obs = Observation()

        if results == []:
            obs.set_success(False)
            return obs
        else:
            obs.add_raw_obs(results)
            obs.set_success(True)

        for p in results:
            obs.add_process(
                hostid=client_id,
                pid=p['Pid'],
                process_name=p['Name'],
                parent_pid=p['Ppid'],
                username=p['Username']
            )

        return obs



# Copyright DST Group. Licensed under the MIT license.

from csle_cyborg.shared.enums import QueryType
from csle_cyborg.shared.observation import Observation
from .velociraptor_action import VelociraptorAction


class GetUserNTDS(VelociraptorAction):
    """Get the NT Directory Services file"""

    def __init__(self, session: int, agent: str, hostname: str, username: str):
        super().__init__(session=session,
                         query_type=QueryType.ASYNC,
                         tag=agent)
        self.agent = agent
        self.hostname = hostname
        self.username = username

    def sim_execute(self, state):
        obs = Observation()
        return obs

    def emu_execute(self,
                    session_handler,
                    *args,
                    **kwargs):
        """Execute and action in emulator environment

        Parameters
        ----------
        session_handler : SessionHandler
           session handler object for game session of action (i.e. that
           matches session_id)

        Returns
        -------
        Observation
            Result of performing action
        """
        raise NotImplementedError



# Copyright DST Group. Licensed under the MIT license.
import time

from csle_cyborg.shared.observation import Observation
from csle_cyborg.shared.actions.agent_actions.agent_action import AgentAction


class AgentSleep(AgentAction):

    def __init__(self, t: int = 1):
        super().__init__()
        self.t = t

    def sim_execute(self, state):
        return Observation()

    def emu_execute(self, agent, *args, **kwargs) -> Observation:
        time.sleep(self.t)
        obs = Observation()
        obs.set_success(True)
        obs.add_raw_obs(f"I'm feeling refreshed! I slept {self.t} secs")
        return obs



from random import choice
from csle_cyborg.shared.observation import Observation
from csle_cyborg.shared.actions.action import Action
from csle_cyborg.shared.enums import DecoyType, OperatingSystemType
from csle_cyborg.simulator.host import Host
from csle_cyborg.simulator.session import Session
from csle_cyborg.simulator.state import State
from csle_cyborg.shared.actions.abstract_actions.misinform import Decoy, DecoyFactory, _is_host_using_port

class SvchostDecoyFactory(DecoyFactory):
    """
    Assembles process information to appear as svchost
    """
    def make_decoy(self, host: Host) -> Decoy:
        del host
        return Decoy(service_name="svchost", name="Svchost.exe",
                open_ports=[{'local_port':3389, 'local_address':'0.0.0.0'}],
                process_type="rdp")

    def is_host_compatible(self, host: Host) -> bool:
        has_port = not _is_host_using_port(host, 3389)
        is_windows = host.os_type == OperatingSystemType.WINDOWS
        return all((has_port,is_windows))
svchost_decoy_factory = SvchostDecoyFactory()


class DecoySvchost(Action):
    """
    Creates a misleading process on the designated host depending on
    available and compatible options.
    """
    def __init__(self, *, session: int, agent: str, hostname: str):
        self.agent = agent
        self.session = session
        self.hostname = hostname
        self.decoy_type = DecoyType.EXPLOIT
        self.candidate_decoys = (svchost_decoy_factory,)

    def emu_execute(self) -> Observation:
        raise NotImplementedError

    def sim_execute(self, state: State) -> Observation:
        obs_fail = Observation(False)
        obs_succeed = Observation(True)

        sessions = [s for s in state.sessions[self.agent].values() if
                s.host == self.hostname]
        if len(sessions) == 0:
            return obs_fail

        session = choice(sessions)
        host = state.hosts[self.hostname]

        try:
            decoy_factory = self.__select_one_factory(host)
            decoy = decoy_factory.make_decoy(host)
            self.__create_process(obs_succeed, session, host, decoy)
            #print ("Misinform Success. Result: {}".format(result))

            return obs_succeed

        except RuntimeError:
            #print ("Misinform FAILURE")
            return obs_fail


    def __select_one_factory(self, host: Host) -> DecoyFactory:
        """
        Examines all decoy factories and returns one randomly compatible one.
        Raises RuntimeError if no compatible ones are found.
        """

        compatible_factories = [factory for factory in self.candidate_decoys
                if factory.is_host_compatible(host) ]

        if len(compatible_factories) == 0:
            raise RuntimeError("No compatible factory")

        return choice(list(compatible_factories))

    def __create_process(self, obs: Observation, sess: Session, host: Host,
            decoy: Decoy) -> None:
        """
        Creates a process & service from Decoy on current host, adds it
        to the observation.
        """

        parent_pid = 1

        process_name = decoy.name
        username = sess.username
        version = decoy.version
        open_ports = decoy.open_ports
        process_type = decoy.process_type
        process_props = decoy.properties

        service_name = decoy.service_name

        new_proc = host.add_process(name=process_name, ppid=parent_pid,
                user=username, version=version, process_type=process_type,
                open_ports=open_ports, decoy_type=self.decoy_type,
                properties=process_props)

        host.add_service(service_name=service_name, process=new_proc.pid,
                session=sess)

        obs.add_process(hostid=self.hostname, pid=new_proc.pid,
                parent_pid=parent_pid, name=process_name,
                username=username, service_name=service_name,
                properties=process_props)

    def __str__(self):
        return f"{self.__class__.__name__} {self.hostname}"



# Copyright DST Group. Licensed under the MIT license.
from ipaddress import IPv4Address, IPv4Network
from csle_cyborg.shared.actions.msf_actions_folder.msf_scanner_folder.msf_scanner import MSFScanner
from csle_cyborg.shared.enums import SessionType
from csle_cyborg.shared.observation import Observation
from csle_cyborg.simulator.state import State


# msf module is post/multi/gather/ping_sweep
class MSFPingsweep(MSFScanner):
    def __init__(self, subnet: IPv4Network, session: int, agent: str, target_session: int):
        super().__init__(session, agent)
        self.subnet = subnet
        self.target_session = target_session
        self.lo = IPv4Address("127.0.0.1")

    def sim_execute(self, state: State):
        obs = Observation()
        if self.session not in state.sessions[self.agent]:
            obs.set_success(False)
            return obs
        from_host = state.sessions['Red'][self.session].host
        session = state.sessions['Red'][self.session]

        if session.session_type != SessionType.MSF_SERVER or not session.active:
            obs.set_success(False)
            return obs

        if self.target_session in state.sessions['Red']:
            target_session = state.sessions['Red'][self.target_session]
        else:
            obs.set_success(False)
            return obs

        if not (
                target_session.session_type == SessionType.METERPRETER or target_session.session_type == SessionType.MSF_SHELL) or not target_session.active:
            obs.set_success(False)
            return obs

        target_session, from_interface = self.get_local_source_interface(local_session=target_session,
                                                                         remote_address=self.subnet.network_address,
                                                                         state=state)

        if from_interface is None:
            obs.set_success(False)
            return obs
        target_hosts = []
        for host in state.subnets[self.subnet].ip_addresses:
            obs.set_success(True)
            target_hosts.append(host)
            obs.add_interface_info(hostid=str(host), ip_address=host, subnet=self.subnet)

        return obs

    def emu_execute(self, session_handler) -> Observation:
        obs = Observation()
        from csle_cyborg.CybORG import MSFSessionHandler
        if type(session_handler) is not MSFSessionHandler:
            obs.set_success(False)
            return obs
        output = session_handler.execute_module(mtype='post', mname='multi/gather/ping_sweep',
                                                opts={'RHOSTS': str(self.subnet), 'SESSION': self.target_session})
        obs.add_raw_obs(output)
        '''[*] Performing ping sweep for IP range 10.0.2.0/23
            [+] 	10.0.2.1 host found
            [+] 	10.0.2.164 host found
            [*] Post module execution completed'''
        if output.find("Post failed") != -1:
            obs.set_success(False)
            return obs
        for line in output.split('\n'):
            if 'host found' in line:
                obs.set_success(True)
                ip_address = line.split(' ')[1].replace('\t', '')
                # session_handler._log_debug(f"New IP Address found: {ip_address}")
                obs.add_interface_info(hostid=str(ip_address), ip_address=ip_address, subnet=self.subnet)
        # session_handler._log_debug(output)
        return obs

    def __str__(self):
        return super(MSFPingsweep, self).__str__() + f", Subnet: {self.subnet}, Client Session: {self.target_session}"



# Copyright DST Group. Licensed under the MIT license.
from ipaddress import IPv4Address
from csle_cyborg.shared.actions.shell_actions_folder.open_connection_folder.credential_access_folder.credential_access import CredentialAccess
from csle_cyborg.simulator.session import SessionType
from csle_cyborg.shared.enums import ProcessType
from csle_cyborg.shared.observation import Observation


class SSHAccess(CredentialAccess):
    def __init__(self, session, agent, username, password, ip_address, port):
        super().__init__(session, agent)
        self.username = username
        self.password = password
        self.target = ip_address
        self.port = port

    def sim_execute(self, state):
        obs = Observation()
        obs.set_success(False)
        if self.session not in state.sessions[self.agent]:
            return obs
        session = state.sessions[self.agent][self.session]

        if not session.active:
            return obs

        if session.session_type != SessionType.SHELL and session.session_type != SessionType.MSF_SHELL and session.session_type != SessionType.SSH:
            return obs

        # find shared subnet of the two hosts
        server_address = None
        target_host = None
        if self.target == IPv4Address("127.0.0.1"):
            server_address = IPv4Address("127.0.0.1")
            target_host = session.host
        else:
            for interface in session.host.interfaces:
                if self.target in interface.subnet.ip_addresses:
                    server_address = interface.ip_address
                    target_host = state.hosts[state.ip_addresses[self.target]]
                    break

        if server_address is None:
            return obs

        obs.add_interface_info(hostid="1", ip_address=str(self.target))

        ssh_proc = None
        port = None
        # should also check for a non-ssh process listening on port 22 - only extra info in obs will be that this
        # process exists on that port and IP
        for proc in target_host.processes:
            if proc.process_type == ProcessType.SSH:
                for conn in proc.connections:
                    if conn['local_port'] == self.port:
                        ssh_proc = proc
                        port = conn
                    break

        if ssh_proc is None or (
                port["local_address"] != IPv4Address("0.0.0.0") and port["local_address"] != self.target):

            return obs

        obs.add_process(hostid="1", local_address=str(self.target), local_port=22, status="open", app_protocol="ssh",
                        process_type="ssh")
        user_found = False
        # should really check all username password combinations (in order) in whatever file is used rather than just
        # user-user
        for user in target_host.users:
            if user.username == self.username and user.password == self.password:
                user_found = True
                break

        if user_found:
            obs.set_success(True)

            new_session = state.add_session(host=target_host.hostname, agent=self.agent,
                                            user=self.username, session_type="msf shell", parent=session)
            session.active = False
            process = new_session.process
            remote_port = session.host.get_ephemeral_port()
            new_connection = {"local_port": self.port,
                              "Application Protocol": "tcp",
                              "remote_address": server_address,
                              "remote_port": remote_port,
                              "local_address": self.target}
            process.connections.append(new_connection)

            remote_port_dict = {'local_port': remote_port,
                                "Application Protocol": "ssh",
                                "local_address": server_address,
                                "remote_address": self.target,
                                "remote_port": 22
                                }
            session.process.connections.append(remote_port_dict)

            obs.add_session_info(hostid="1", username=self.username, session_id=new_session.ident, session_type="ssh",
                                 timeout=0, agent=self.agent)
            obs.add_user_info(hostid="1", username=self.username, password=self.password)
            obs.add_system_info(hostid='1', os_type=target_host.os_type, hostname=target_host.hostname)

        return obs



# Copyright DST Group. Licensed under the MIT license.
from csle_cyborg.shared.actions.shell_actions_folder.shell_action import ShellAction
from csle_cyborg.shared.enums import OperatingSystemType
from csle_cyborg.shared.observation import Observation


class DeleteFileWindows(ShellAction):
    def __init__(self, session: int, agent: str, file: str, path: str):
        super().__init__(session, agent)
        self.file = file
        self.path = path

    def sim_execute(self, state):
        obs = Observation()
        obs.set_success(False)
        if self.session not in state.sessions[self.agent]:
            return obs
        if not state.sessions[self.agent][self.session].active:
            return obs

        host = state.sessions[self.agent][self.session].host
        obs.add_system_info(hostid="hostid0", os_type=host.os_type)
        if host.os_type == OperatingSystemType.WINDOWS:
            file = host.get_file(name=self.file, path=self.path)
            if file is not None:
                obs.set_success(True)
                host.files.remove(file)

            else:
                obs.set_success(False)

        else:
            obs.set_success(False)
        return obs



# Copyright DST Group. Licensed under the MIT license.
import sys
import logging
import paramiko
import os.path as osp


class CybORGLogger:
    """A logger class for CybORG.

    It has two main functions:
    1. acts as a wrapper for the Python logger class
    2. provides a base class with useful logging function that other classes
    can inherit and use to make logging easier.
    """

    logger_name = "CybORGLog-Process"
    sshtunnel_logger_name = f"{logger_name}-sshtunnel"

    # Add extra levels to logging
    DEBUG2 = "DEBUG2"
    DEBUG2_LVL = logging.DEBUG - 1
    logging.addLevelName(DEBUG2_LVL, DEBUG2)

    @staticmethod
    def setup(config, verbosity: int = None):
        """Setup the CybORG logger using given configuration.

        Arguments
        ---------
        config : CybORGConfig
            the configuration object
        verbosity : int, optional
            verbosity level of console logger, if None uses level in config.
            Level 0 = logging.WARNING (30) and above
            Level 1 = logging.INFO (20) and above
            Level 2 = logging.WARNING (10) and above
            Level 3 = CybORGLogger.DEBUG2 (9) and above (i.e. will show
                      messages logged with the debug2() method.
            Level 4+ = logging.NOTSET (0) and above (i.e. will display all
                       logged information)
        """
        console_log_level = config.default_console_log_level
        if verbosity:
            assert verbosity >= 0, "Invalid verbosity, must be >= 0"
            if verbosity <= 2:
                console_log_level = logging.WARNING - verbosity * 10
            elif verbosity == 3:
                console_log_level = CybORGLogger.DEBUG2
            else:
                console_log_level = logging.NOTSET

        CybORGLogger.logger_name = config.logger_name
        formatter = logging.Formatter(
            fmt=config.logging_format, datefmt=config.logging_date_format
        )
        console_log_level = logging.getLevelName(console_log_level)
        file_log_level = logging.getLevelName(config.default_file_log_level)
        logger = logging.getLogger(config.logger_name)

        # create console handler
        ch = logging.StreamHandler(sys.stdout)
        ch.setLevel(console_log_level)
        ch.setFormatter(formatter)
        logger.addHandler(ch)

        # set the log level of the logger itself
        logger.setLevel(console_log_level)

        # Do NOT propogate log messages to the root logger
        logger.propagate = False

        if config.log_to_file:
            log_file = osp.join(config.log_dir_path, config.logger_file_name)
            fh = logging.FileHandler(filename=log_file, mode='w')
            fh.setLevel(file_log_level)
            fh.setFormatter(formatter)
            logger.addHandler(fh)
            logger.setLevel(file_log_level)

        ##################################################################
        # Paramiko logging config
        ##################################################################
        paramiko_logger_name = f"{config.logger_name}-paramiko"
        paramiko_logger = paramiko.util.get_logger("paramiko")
        # Suppress paramiko's verbose output to stdout
        paramiko_log_file = osp.join(
            config.log_dir_path, paramiko_logger_name + ".log"
        )
        paramiko.util.log_to_file(paramiko_log_file, level="WARN")
        paramiko_logger.setLevel(logging.WARNING)
        paramiko_logger.propagate = False  # don't dump on console

        ##################################################################
        # SSHTunnel logging config
        ##################################################################
        ssht_log_format = "%(asctime)-15s (%(levelname)-8s) ==> %(message)s"
        ssht_formatter = logging.Formatter(
            fmt=ssht_log_format, datefmt=config.logging_date_format
        )

        # note this is not the default logger name for SSHTunnel,
        # just a name chosen for CybORG
        sshtunnel_logger_name = f"{config.logger_name}-sshtunnel"
        sshtunnel_logger = logging.getLogger(sshtunnel_logger_name)
        # Don't dump sshtunnel outputs to console
        sshtunnel_logger.propagate = False
        if config.log_to_file:
            # create file handler
            ssh_log_file = osp.join(
                config.log_dir_path, sshtunnel_logger_name + ".txt"
            )
            sfh = logging.FileHandler(ssh_log_file, mode='w')
            sfh.setLevel(file_log_level)
            # create formatter and add it to the handlers
            sfh.setFormatter(ssht_formatter)
            # add the handlers to the sshtunnel_logger
            sshtunnel_logger.addHandler(sfh)

    @staticmethod
    def setLevel(level):
        logging.getLogger(
            CybORGLogger.logger_name
        ).setLevel(level=level)

    @staticmethod
    def debug(msg, *args, **kwargs):
        logging.getLogger(
            CybORGLogger.logger_name
        ).debug(msg, *args, **kwargs)

    @staticmethod
    def debug2(msg, *args, **kwargs):
        logging.getLogger(
            CybORGLogger.logger_name
        ).log(CybORGLogger.DEBUG2_LVL, msg, *args, **kwargs)

    @staticmethod
    def info(msg, *args, **kwargs):
        logging.getLogger(
            CybORGLogger.logger_name
        ).info(msg, *args, **kwargs)

    @staticmethod
    def warning(msg, *args, **kwargs):
        logging.getLogger(
            CybORGLogger.logger_name
        ).warning(msg, *args, **kwargs)

    @staticmethod
    def error(msg, *args, **kwargs):
        logging.getLogger(
            CybORGLogger.logger_name
        ).error(msg, *args, **kwargs)

    @staticmethod
    def critical(msg, *args, **kwargs):
        logging.getLogger(
            CybORGLogger.logger_name
        ).critical(msg, *args, **kwargs)

    @staticmethod
    def header(title):
        CybORGLogger.info(f"\n\n{'':*^30} {title:^50} {'':*^30}\n\n")

    @staticmethod
    def get_logger():
        return logging.getLogger(CybORGLogger.logger_name)

    @staticmethod
    def get_ssh_tunnel_logger():
        return logging.getLogger(CybORGLogger.sshtunnel_logger_name)

    def _log_header(self, title):
        CybORGLogger.header(self._format_log_msg(title))

    def _log_info(self, msg):
        CybORGLogger.info(self._format_log_msg(msg))

    def _log_error(self, msg):
        CybORGLogger.error(self._format_log_msg(msg))

    def _log_debug(self, msg):
        CybORGLogger.debug(self._format_log_msg(msg))

    def _log_debug2(self, msg):
        CybORGLogger.debug2(self._format_log_msg(msg))

    def _log_warning(self, msg):
        CybORGLogger.warning(self._format_log_msg(msg))

    def _format_log_msg(self, msg):
        """Overide this function for more informative logging messages """
        return f"{self.__class__.__name__}: {msg}"


def log_trace(func):
    """Logger decorator for logging function execution.

    Import this function and add @log_trace above your function of
    interest to log output to file about the functions execution
    """

    def call(*args, **kwargs):
        """ Actual wrapping """
        entering(func, *args)
        result = func(*args, **kwargs)
        exiting(func)
        return result

    return call


def entering(func, *args):
    """ Pre function logging """
    CybORGLogger.debug("Entered %s", func.__name__)
    CybORGLogger.debug(func.__doc__)
    CybORGLogger.debug(
        "Function at line %d in %s" % (func.__code__.co_firstlineno,
                                       func.__code__.co_filename)
    )

    try:
        CybORGLogger.debug(
            "The argument %s is %s" % (func.__code__.co_varnames[0], *args)
        )
    except IndexError:
        CybORGLogger.debug("No arguments")


def exiting(func):
    """ Post function logging """
    CybORGLogger.debug("Exited  %s", func.__name__)



import os
import errno
import os.path as osp
from configparser import ConfigParser


def get_parent_dir_path(path):
    """Returns the path to the parent directory in path. """
    if osp.isfile(path):
        return osp.dirname(path)
    return osp.dirname(osp.abspath(path))


# Default file names
CONFIG_FILE_NAME = "config.ini"
DEFAULT_VALUES_FILE_NAME = "defaultconfig.ini"


# Default config dir = dir this file is in
DEFAULT_CONFIG_DIR = osp.dirname(osp.abspath(__file__))

# DEFAULT CYBORG DIR = parent dir of config dir
DEFAULT_CYBORG_DIR = get_parent_dir_path(
    get_parent_dir_path(DEFAULT_CONFIG_DIR)
)

# DEFAULT_CONFIG_FILE_PATH
DEFAULT_CONFIG_FILE_PATH = osp.join(DEFAULT_CONFIG_DIR, CONFIG_FILE_NAME)

# Keys for sections of config file
DIRS = "DIRS"
LOGGING = "LOGGING"

# Keys for config variables
CYBORG_BASE_DIR = "cyborg_base_dir"
LOGGER_NAME = "logger_name"
LOG_FILE = "log_file"
LOG_DIR_PATH = "log_dir_path"
DEFAULT_CONSOLE_LOG_LEVEL = "default_console_log_level"
DEFAULT_FILE_LOG_LEVEL = "default_file_log_level"

SECTION_MAP = {
    DIRS: [CYBORG_BASE_DIR],
    LOGGING: [
        LOGGER_NAME,
        LOG_FILE,
        LOG_DIR_PATH,
        DEFAULT_CONSOLE_LOG_LEVEL,
        DEFAULT_FILE_LOG_LEVEL
    ]
}

INI_FILETYPES = (("ini files", "*.ini"), ("all files", "*.*"))
KEY_FILETYPES = (("key files", "*.pem"), ("all files", "*.*"))


def config_file_valid(config_file_path: str, section_map: dict) -> bool:
    """Checks to see if config file exists and is valid

    Parameters
    ----------
    config_file : str
        Path to config file to check
    section_map : dict
        dictionary of expected sections and variables within config file

    Returns
    -------
    bool
        True if config file exists and is valid
    """
    if not osp.isfile(config_file_path):
        return False

    try:
        config = ConfigParser()
        config.read(config_file_path)

        if not validate_config_file(config, section_map):
            return False

        file_dirs = [config[DIRS][CYBORG_BASE_DIR],
                     config[LOGGING][LOG_DIR_PATH]]
        if not validate_file_dirs(file_dirs):
            return False
        return True

    except Exception:
        return False


def validate_config_file(config: ConfigParser, section_map: dict) -> bool:
    """Check that config file is valid and contains all expected variables.

    Parameters
    ----------
    config: ConfigParser
        Config to validate
    section_map : dict
        dictionary of expected sections and variables within config file

    Returns
    -------
    bool
        True if config file is valid
    """
    # Check all the variables are there
    for section, config_vars in section_map.items():
        for var in config_vars:
            if var not in config[section]:
                print(f"Config file invalid, variable '{var}' missing"
                      f" from '{section}' section")
                return False
    return True


def validate_file_dirs(file_dirs: list) -> bool:
    """Check that all directories designated are valid.

    Parameters
    ----------
    file_dirs: list
        list of directory paths

    Returns
    -------
    bool
        True if all directory paths are valid.
    """
    for f in file_dirs:
        if not osp.isdir(f):
            return False
    return True


def file_path_valid(file_path: str) -> bool:
    """Check file path is a valid file.

    Parameters
    ----------
    file_path : str
        the file path

    Returns
    -------
    bool
        True if the file path is valid
    """
    return osp.isfile(file_path)


def validate_file_paths(file_paths: list) -> bool:
    """Check that all files paths in config file are valid.

    Parameters
    ----------
    file_paths : list
        list of file paths

    Returns
    -------
    bool
        True if all file paths are valid
    """
    for fp in file_paths:
        if not osp.isfile(fp):
            return False
    return True


def remove_file_if_exists(filepath: str):
    """Remove a file at location if it exists. """
    try:
        os.remove(filepath)
    except OSError as e:
        if e.errno != errno.ENOENT:
            # errno.ENOENT = no such file or directory
            # re-raise exception if a different error occurred
            raise e


def ask_yes_no_question(question: str) -> bool:
    """Ask user a yes/no question in terminal.

    Parameters
    ----------
    question: str
        The question to ask the user

    Returns
    -------
    bool
        True if the user answered 'yes', False if user answered 'no'
    """
    while True:
        yes_no = input(f"{question} [y|n] ")
        if yes_no.strip().lower() == "y":
            return True
        elif yes_no.strip().lower() == "n":
            return False
        else:
            print("\tPlease just type 'y' or 'n'")


def ask_file_path(question: str) -> str:
    """Ask user to provide the path to a file.

    Parameters
    ----------
    question: str
        The question to ask the user

    Returns
    -------
    str
       the user supplied file path
    """
    while True:
        fp = input(f"{question}:")
        if file_path_valid(fp):
            return fp
        print("Invalid file path. Try again.")


def ask_file_dir(question: str) -> str:
    """Ask user to provide the path to a directory.

    Parameters
    ----------
    question: str
        The question to ask the user

    Returns
    -------
    str
       the user supplied file path
    """
    while True:
        fp = input(f"{question}:")
        if osp.isdir(fp):
            return fp
        print("Invalid directory path. Try again.")


class ConfigHelperGUI:
    """Class for running Config creation GUI. """

    def __init__(self):
        self.init_dir = DEFAULT_CONFIG_DIR
        self.default_config_file_path = DEFAULT_CONFIG_FILE_PATH
        self.default_values_file_name = DEFAULT_VALUES_FILE_NAME
        self.config_file_name = CONFIG_FILE_NAME
        self.default_cyborg_dir = DEFAULT_CYBORG_DIR

    def run_helper_gui(self):
        """Run Config Helper GUI.

        Returns
        -------
        config_valid : bool
            whether valid config exists
        config_file_path : str
            path to config file, if it exists
        """
        try:
            if not self._user_wants_to_create_new_file():
                return False, ""

            self.config_file_path = self._get_config_file_path()
            if config_file_valid(self.config_file_path, SECTION_MAP):
                if self._use_existing_file():
                    return True, self.config_file_path
                remove_file_if_exists(self.config_file_path)

            self.default_config = self._load_default_config()
            new_config = self._create_new_config()
            self._write_new_config(new_config, self.config_file_path)
            return True, self.config_file_path

        except Exception as ex:
            print(f"Exception creating config file: \n{str(ex)}")
            return False, ""

    def _user_wants_to_create_new_file(self):
        question = (f"ERROR: The config file '{self.default_config_file_path}'"
                    "does not exist or is invalid. Would you like to create "
                    " one now?")
        if not ask_yes_no_question(question):
            print("Can't run CybORG. Please create config file and/or "
                  "check paths and permissions.")
            return False
        return True

    def _get_config_file_path(self):
        print(f"Creating config file at {self.default_config_file_path}")
        return osp.abspath(self.default_config_file_path)

    def _use_existing_file(self):
        question = (f"The config file \"{self.config_file_path}\" exists. "
                    "Would you like to use it (say no to create new one)?")
        if ask_yes_no_question(question):
            return True
        return False

    def _load_default_config(self):
        default_config_path = osp.join(self.init_dir,
                                       self.default_values_file_name)
        if not osp.isfile(default_config_path):
            title = "Please locate the default config file"
            fp = ask_file_path(title)
            default_config_path = osp.abspath(fp)

        config_parse = ConfigParser()
        config_parse.read(default_config_path)
        return config_parse

    def _create_new_config(self):
        print(f"Creating new config file...")
        new_config = ConfigParser()

        self._config_dirs(new_config)
        self._config_logging(new_config)
        return new_config

    def _config_dirs(self, new_config):
        print(f"\nConfiguration for CybORG Shared: {DIRS}")
        new_config[DIRS] = {}
        for var in SECTION_MAP[DIRS]:
            if var == CYBORG_BASE_DIR:
                self.base_dir = self._get_base_dir()
                new_config[DIRS][var] = self.base_dir
            else:
                value = self._get_config_value(DIRS, var)
                new_config[DIRS][var] = value

    def _get_base_dir(self):
        base_dir = self.default_cyborg_dir
        question = f"CybORG source directory is {base_dir}, is that correct?"
        if not ask_yes_no_question(question):
            title = "Please locate the CybORG source directory"
            base_dir = ask_file_dir(title)

        if osp.isdir(base_dir):
            return base_dir

        raise Exception("Create config file failed, the base directory of the "
                        f" repository ({base_dir}), is not  valid.")

    def _config_logging(self, new_config):
        print(f"\nConfiguration for CybORG Shared: {LOGGING}")
        new_config[LOGGING] = {}
        for var in SECTION_MAP[LOGGING]:
            if var == LOG_DIR_PATH:
                log_dir = self._get_log_dir()
                new_config[LOGGING][var] = log_dir
            else:
                value = self._get_config_value(LOGGING, var)
                new_config[LOGGING][var] = value

    def _get_log_dir(self):
        log_dir = osp.join(self.base_dir,
                           self.default_config[LOGGING][LOG_DIR_PATH])
        question = (f"Logs will be written to directory ({log_dir}), "
                    "do you wish to specify a different directory?")
        if ask_yes_no_question(question):
            title = "Please choose a log directory"
            log_dir = ask_file_dir(title)

        if not osp.isdir(log_dir):
            os.makedirs(log_dir)
        return log_dir

    def _get_config_value(self, section, var):
        default = self.default_config[section][var]
        if isinstance(default, bool):
            input_var = input(f"Do you want to {var} (default={default}):")
        else:
            input_var = input(f"Please enter {var} (default={default}):")
        if not input_var:
            input_var = default
        return input_var

    def _write_new_config(self, new_config, config_file_path):
        with open(config_file_path, 'w') as newini:
            new_config.write(newini)


if __name__ == "__main__":
    ConfigHelperGUI().run_helper_gui()



## The following code contains work of the United States Government and is not subject to domestic copyright protection under 17 USC § 105.
## Additionally, we waive copyright and related rights in the utilized code worldwide through the CC0 1.0 Universal public domain dedication.
from ipaddress import IPv4Address
from typing import Optional
from csle_cyborg.shared.actions.concrete_actions.exploit_action import ExploitAction
from csle_cyborg.shared.actions.msf_actions_folder.msf_action import lo
from csle_cyborg.shared.enums import ProcessType, DecoyType
from csle_cyborg.shared.observation import Observation
from csle_cyborg.simulator.host import Host
from csle_cyborg.simulator.state import State
from csle_cyborg.simulator.process import Process


class SSHBruteForce(ExploitAction):
    def __init__(self, ip_address: IPv4Address, agent: str, session: int, target_session: int):
        super().__init__(session=session, agent=agent, ip_address=ip_address,
                         target_session=target_session)
        self.ip_address = ip_address
        self.target_session = target_session

    def sim_execute(self, state: State):
        self.state = state
        length_of_wordlist = 10
        obs = Observation()
        if self.session not in state.sessions[self.agent]:
            obs.set_success(False)
            return obs
        from_host: Host = state.hosts[state.sessions['Red'][self.session].host]
        session = state.sessions['Red'][self.session]

        if not session.active:
            obs.set_success(False)
            return obs

        # determine which ports can communicate between which subnets
        originating_ip_address = None
        if self.ip_address == lo:
            target_host: Host = from_host
            originating_ip_address = self.ip_address
        else:
            target_host: Host = state.hosts[state.ip_addresses[self.ip_address]]
            ports = self.check_routable(
                [state.subnets[i.subnet] for i in from_host.interfaces if i.ip_address != lo],
                [s for s in state.subnets.values() if self.ip_address in s.cidr])
            if ports is None or (22 not in ports and 'all' not in ports):
                obs.set_success(False)
                return obs
            from_subnet, to_subnet = ports[22] if 22 in ports else ports['all']
            # calculate the originating ip address
            for i in from_host.interfaces:
                if i.ip_address != lo:
                    if i.subnet == from_subnet:
                        originating_ip_address = i.ip_address

        # find out if smb is open
        vuln_proc: Optional[Process] = None
        for proc in target_host.processes:
            if proc.process_type == ProcessType.SSH:
                for conn in proc.connections:
                    if 'local_port' in conn and conn['local_port'] == 22:
                        vuln_proc = proc
                        break
                if vuln_proc is not None:
                    break

        if vuln_proc is None:
            obs.set_success(False)
            return obs
        obs.add_process(hostid=str(self.ip_address), local_address=self.ip_address, local_port=22, status="open",
                        process_type='SSH')

        # test if there is a bruteforceable user-pass on the system
        user = None
        for u in target_host.users:
            if u.bruteforceable:
                user = u
                break

        for i in range(length_of_wordlist):
            target_host.events['NetworkConnections'].append({'remote_address': originating_ip_address,
                                                             'remote_port': from_host.get_ephemeral_port(),
                                                             'local_address': self.ip_address,
                                                             'local_port': 22
                                                             })

        if user is not None and not (vuln_proc.decoy_type & DecoyType.EXPLOIT):
            obs.set_success(True)

            new_proc = target_host.add_process(name="sshd", ppid=vuln_proc.pid, path=vuln_proc.path, user=user.username,
                                               process_type="ssh")

            if bool(vuln_proc.decoy_type & DecoyType.SANDBOXING_EXPLOIT):

                new_session = state.add_session(host=target_host.hostname, agent=self.agent,
                                                user=user.username, session_type="ssh", parent=session,
                                                process=new_proc.pid,
                                                is_escalate_sandbox=True)
            else:

                new_session = state.add_session(host=target_host.hostname, agent=self.agent,
                                                user=user.username, session_type="ssh", parent=session,
                                                process=new_proc.pid)

            remote_port = target_host.get_ephemeral_port()
            new_connection = {"local_port": 22,
                              "Application Protocol": "tcp",
                              "remote_address": originating_ip_address,
                              "remote_port": remote_port,
                              "local_address": self.ip_address}
            new_proc.connections.append(new_connection)
            target_host.events['NetworkConnections'].append({'remote_address': originating_ip_address,
                                                             'remote_port': remote_port,
                                                             'local_address': self.ip_address,
                                                             'local_port': 22
                                                             })

            remote_port_dict = {'local_port': remote_port,
                                "Application Protocol": "ssh",
                                "local_address": originating_ip_address,
                                "remote_address": self.ip_address,
                                "remote_port": 22
                                }
            from_host.get_process(session.pid).connections.append(remote_port_dict)
            obs.add_process(hostid=str(originating_ip_address), local_address=originating_ip_address,
                            remote_address=self.ip_address,
                            local_port=remote_port, remote_port=22)
            obs.add_process(hostid=str(self.ip_address), local_address=self.ip_address,
                            remote_address=originating_ip_address,
                            local_port=22, remote_port=remote_port, process_type='ssh')
            obs.add_session_info(hostid=str(self.ip_address), username=user.username, session_id=new_session.ident,
                                 session_type="ssh", agent=self.agent)
            obs.add_user_info(hostid=str(self.ip_address), username=user.username, password=user.password, uid=user.uid)

            obs.add_system_info(hostid=str(self.ip_address), hostname=target_host.hostname, os_type=target_host.os_type)

        else:
            obs.set_success(False)
        return obs



# Copyright DST Group. Licensed under the MIT license.
from ipaddress import IPv4Address
from time import sleep

from csle_cyborg.shared.actions.msf_actions_folder.remote_code_execution_folder.remote_code_execution import RemoteCodeExecution
# use msf module exploit/windows/smb/ms17_010_eternal_blue, set RHOSTS to target
# could also change LHOST, LPORT and RPORT (default 139)
# gives root session
from csle_cyborg.simulator.session import SessionType
from csle_cyborg.shared.enums import OperatingSystemType
from csle_cyborg.shared.observation import Observation
from csle_cyborg.simulator.user import User


class MS17_010_PSExec(RemoteCodeExecution):
    def __init__(self, ip_address: IPv4Address, session: int, agent: str, username: str, password: str):
        super().__init__(session=session, agent=agent)
        self.target = ip_address
        self.username = username
        self.password = password
        self.port = 139

    def sim_execute(self, state):
        obs = Observation()
        obs.set_success(False)
        if self.session not in state.sessions[self.agent]:
            return obs
        session = state.sessions[self.agent][self.session]

        if session.session_type != SessionType.MSF_SERVER or not session.active:
            return obs
        target_subnet = None
        for subnet in state.subnets.values():
            if self.target in subnet.ip_addresses:
                target_subnet = subnet
                break
        if target_subnet is None:
            return obs
        # find shared subnet of the two hosts
        server_session, server_interface = self.get_local_source_interface(local_session=session,
                                                                           remote_address=self.target, state=state)

        if server_interface is None:
            return obs

        server_address = server_interface.ip_address

        if self.target == IPv4Address("127.0.0.1"):
            target_host = server_session.host
        else:
            target_host = state.hosts[state.ip_addresses[self.target]]
            if not self.test_nacl(port=self.port, target_subnet=target_subnet,
                                  originating_subnet=state.subnets[server_interface.subnet]):
                return obs

        # obs.add_interface_info(hostid='0', ip_address=server_address)
        # obs.add_interface_info(hostid='1', ip_address=self.target)

        # find out if smb is open
        smb_proc = None
        for proc in target_host.processes:
            for conn in proc.connections:
                if conn['local_port'] == self.port and 'remote_address' not in conn:
                    # if proc.process_type == ProcessType.SMB:
                    # TODO: In case of SMB that is not the right version, should SMB process be in the obs?
                    smb_proc = proc
                    break

        # find out if smb is vulnerable (Windows OS + smb version)
        # Note that this exploit should actually work for all versions in the range Samba 3.0.20 - 3.0.25rc3
        if smb_proc is not None:  # and smb_proc.version == ProcessVersion.SMBv1:
            obs.add_process(hostid=str(self.target), local_address=self.target, local_port=self.port, status="open",
                            process_type="smb")
            if target_host.os_type == OperatingSystemType.WINDOWS:
                obs.set_success(True)

                root_user: User = None
                for u in state.hosts[state.ip_addresses[self.target]].users:
                    if u.username == "SYSTEM":
                        root_user = u

                new_session = state.add_session(host=target_host.hostname, agent=self.agent,
                                                user=root_user.username, session_type="meterpreter",
                                                parent=server_session)

                local_port = target_host.get_ephemeral_port()
                new_connection = {"remote_port": local_port,
                                  "Application Protocol": "tcp",
                                  "remote_address": server_address,
                                  "local_port": 44444,
                                  "local_address": self.target
                                  }
                state.hosts[new_session.host].get_process(new_session.pid).connections.append(new_connection)

                target_host.add_process(name="telnet", ppid=1, path="/usr/bin/", user=root_user)
                remote_port = {"remote_port": 44444,
                               "Application Protocol": "tcp",
                               "local_address": server_address,
                               "remote_address": self.target,
                               "local_port": local_port
                               }

                state.hosts[server_session.host].get_process(server_session.pid).connections.append(remote_port)
                if session != server_session:
                    local_port = None
                obs.add_process(hostid=str(self.target), local_address=str(self.target), remote_address=server_address,
                                remote_port=local_port, local_port=44444)
                obs.add_session_info(hostid=str(self.target), session_id=int(new_session.ident),
                                     session_type=new_session.session_type, agent=self.agent)
            else:
                obs.add_interface_info(ip_address=str(self.target))

        return obs

    def emu_execute(self, session_handler) -> Observation:
        raise ValueError("not supported")

    def __str__(self):
        return super(MS17_010_PSExec,
                     self).__str__() + f", Target: {self.target}, Username: {self.username}, Password: {self.password}"



from random import choice

from csle_cyborg.shared.observation import Observation
from csle_cyborg.shared.actions.action import Action
from csle_cyborg.shared.actions.concrete_actions.density_scout import DensityScout
from csle_cyborg.shared.actions.concrete_actions.sig_check import SigCheck


class Analyse(Action):
    def __init__(self, session: int, agent: str, hostname: str):
        super().__init__()
        self.agent = agent
        self.session = session
        self.hostname = hostname

    def sim_execute(self, state) -> Observation:
        # perform monitor at start of action
        # monitor = Monitor(session=self.session, agent=self.agent)
        # obs = monitor.sim_execute(state)

        artefacts = [DensityScout, SigCheck]
        # find relevant session on the chosen host
        sessions = [s for s in state.sessions[self.agent].values() if s.host == self.hostname]
        if len(sessions) > 0:
            session = choice(sessions)
            # run the artifacts on the chosen host
            obs = Observation(True)
            for artifact in artefacts:
                sub_action = artifact(agent=self.agent, session=self.session, target_session=session.ident)
                sub_obs = sub_action.sim_execute(state)
                obs.combine_obs(sub_obs)
            return obs
        else:
            return Observation(False)

    def __str__(self):
        return f"{self.__class__.__name__} {self.hostname}"



# Copyright DST Group. Licensed under the MIT license.

import pkgutil
import sys
import inspect


class Plugins:

    def __init__(self, plugin_dir):

        self.loadMenuItemsFromPlugins(self, plugin_dir)

    actions_dict = {"Blue": [], "Red": []}

    @staticmethod
    def getActionNames(self, action_type: str):
        """
        Return a list of all of the actions available for a particular type.
        """
        return [o.name for o in self.actions_dict[action_type]]

    @staticmethod
    def getActionByName(self,  action_type: str, name: str):

        """
        Return an action for a given name for a particular type.
        """
        return self.actions_dict(action_type)[name]


    @staticmethod
    def loadMenuItemsFromPlugins(self, plugin_dir):
        default_plugin_dir = "Shared\Plugins"

        if plugin_dir is None:
            plugin_dir = default_plugin_dir

        # This is from https://stackoverflow.com/questions/1057431/how-to-load-all-modules-in-a-folder

        for importer, package_name, _ in pkgutil.iter_modules([plugin_dir]):
            full_package_name = '%s.%s' % (plugin_dir, package_name)

            if full_package_name not in sys.modules:
                # print("full_package_name=%s\n" % full_package_name)
                # print("package_name=%s\n" % package_name)

                module = importer.find_module(package_name).load_module(full_package_name)

                print("module=%s\n" % str(module))

                members = inspect.getmembers(module)

                for member in members:  # Only add tuple with "CybORGPlugin", but excluding the base class

                    if member[0].find(str("CybORGPlugin")) > -1 and len(member[0]) > len("CybORGPlugin"):
                        plugin_object = member[1]()
                        self.actions_dict[plugin_object.PluginType].append(plugin_object)



from csle_cyborg.agents.simple_agents.base_agent import BaseAgent
from csle_cyborg.shared.results import Results
from csle_cyborg.shared.actions.abstract_actions.monitor import Monitor
from csle_cyborg.shared.actions.abstract_actions.remove import Remove
from csle_cyborg.shared.actions.abstract_actions.restore import Restore



class BlueReactRemoveAgent(BaseAgent):
    def __init__(self):
        self.host_list = []
        self.last_action = None

    def train(self, results: Results):
        pass

    def get_action(self, observation, action_space):
        # add suspicious hosts to the hostlist if monitor found something
        # added line to allow for automatic monitoring.
        if self.last_action is not None and self.last_action == 'Monitor':
            for host_name, host_info in [(value['System info']['Hostname'], value) for key, value in observation.items() if key != 'success']:
                if host_name not in self.host_list and host_name != 'User0' and 'Processes' in host_info and len([i for i in host_info['Processes'] if 'PID' in i]) > 0:
                    self.host_list.append(host_name)
        # assume a single session in the action space
        session = list(action_space['session'].keys())[0]
        if len(self.host_list) == 0:
            self.last_action = 'Monitor'
            return Monitor(agent='Blue', session=session)
        else:
            self.last_action = 'Remove'
            return Remove(hostname=self.host_list.pop(0), agent='Blue', session=session)

    def end_episode(self):
        self.host_list = []
        self.last_action = None

    def set_initial_values(self, action_space, observation):
        pass


class BlueReactRestoreAgent(BaseAgent):
    def __init__(self):
        self.host_list = []
        self.last_action = None

    def train(self, results: Results):
        pass

    def get_action(self, observation, action_space):
        # add suspicious hosts to the hostlist if monitor found something
        # added line to reflect changes in blue actions
        if self.last_action is not None and self.last_action == 'Monitor':
            for host_name, host_info in [(value['System info']['Hostname'], value) for key, value in observation.items() if key != 'success']:
                if host_name not in self.host_list and host_name != 'User0' and 'Processes' in host_info and len([i for i in host_info['Processes'] if 'PID' in i]) > 0:
                    self.host_list.append(host_name)
        # assume a single session in the action space
        session = list(action_space['session'].keys())[0]
        if len(self.host_list) == 0:
            self.last_action = 'Monitor'
            return Monitor(agent='Blue', session=session)
        else:
            self.last_action = 'Restore'
            return Restore(hostname=self.host_list.pop(0), agent='Blue', session=session)

    def end_episode(self):
        self.host_list = []
        self.last_action = None

    def set_initial_values(self, action_space, observation):
        pass



from copy import deepcopy
from prettytable import PrettyTable
from csle_cyborg.agents.wrappers.base_wrapper import BaseWrapper

class TrueTableWrapper(BaseWrapper):
    def __init__(self,env=None,agent=None, observer_mode=True):
        super().__init__(env,agent)
        self.scanned_ips = set()
        self.step_counter = -1
        self.observer_mode = observer_mode

    def reset(self, agent=None):        
        self.scanned_ips = set()  
        self.step_counter = -1
        result = self.env.reset(agent)        
        result.observation = self.observation_change(result.observation)        
        return result      

    def observation_change(self,observation):
        self.step_counter +=1
        self._update_scanned()

        return observation if self.observer_mode else self._create_true_table()

    def get_table(self):
        return self._create_true_table()

    def _update_scanned(self):
        if self.step_counter <= 0:
            return

        action = self.get_last_action(agent='Red')
        if action.__class__.__name__ == 'DiscoverNetworkServices':
            red_obs = deepcopy(self.get_observation(agent='Red'))
            success = red_obs['success']
            if success:
                ip = red_obs.popitem()[0]
                self.scanned_ips.add(ip)

    def _create_true_table(self):
        true_obs = deepcopy(self.env.get_agent_state('True'))
        success = true_obs.pop('success')

        table = PrettyTable([
            'Subnet',
            'IP Address',
            'Hostname',
            'Known',
            'Scanned',
            'Access',
            ])

        for hostid in true_obs:
            host = true_obs[hostid]
            for interface in host['Interface']:
                ip = interface['IP Address']
                if str(ip) == '127.0.0.1':
                    continue
                if 'Subnet' not in interface:
                    continue
                subnet = interface['Subnet']
                hostname = host['System info']['Hostname']
                action_space = self.get_action_space(agent = 'Red')
                known = action_space['ip_address'][ip]
                scanned = True if str(ip) in self.scanned_ips else False
                access = self._determine_red_access(host['Sessions'])

                table.add_row([subnet,str(ip),hostname,known,scanned,access])
        
        table.sortby = 'Hostname'
        table.success = success
        return table

    def _determine_red_access(self,session_list):
        for session in session_list:
            if session['Agent'] != 'Red':
                continue
            privileged = session['Username'] in {'root','SYSTEM'}
            return 'Privileged' if privileged else 'User'

        return 'None'


    def get_attr(self,attribute:str):
        return self.env.get_attr(attribute)

    def get_observation(self, agent: str):
        return self.get_attr('get_observation')(agent)

    def get_agent_state(self,agent:str):
        if agent == 'True':
            output = self.get_table()
        else:
            output = self.get_attr('get_agent_state')(agent)

        return output

    def get_action_space(self,agent):
        return self.get_attr('get_action_space')(agent)

    def get_last_action(self,agent):
        return self.get_attr('get_last_action')(agent)

    def get_ip_map(self):
        return self.get_attr('get_ip_map')()

    def get_rewards(self):
        return self.get_attr('get_rewards')()

def true_obs_to_table(true_obs,env):
    print('Scanned column likely inaccurate.')
    wrapper = TrueTableWrapper(env,observer_mode=False)
    wrapper.step_counter = 1
    return wrapper.observation_change(true_obs)



from csle_cyborg.agents.simple_agents.base_agent import BaseAgent
import random

#agent that does random action

class TestAgent(BaseAgent):

    def __init__(self, test_action=None, epsilon=1.0):
        self.test_action = test_action
        self.epsilon = epsilon

    def train(self, results):
        pass

    def get_action(self, observation: dict, action_space: dict):
        if (random.random() < self.epsilon) or (self.test_action == None):
            action = []
            for a in action_space:
                if a > 0:
                    action.append(random.choice(range(a)))
                else:
                    action.append(0)
            return action
        else:
            return self.test_action

    def end_episode(self):
        pass

    def set_initial_values(self, action_space, observation):
        pass




from csle_cyborg.shared.actions.abstract_actions.monitor import Monitor
from csle_cyborg.agents.simple_agents.base_agent import BaseAgent

class BlueMonitorAgent(BaseAgent):
    def __init__(self):
        pass

    def get_action(self,observation,action_space):
        session = 0
        return Monitor(session=session,agent='Blue')


    def train(self, results):
        pass

    def end_episode(self):
        pass

    def set_initial_values(self, action_space, observation):
        pass



# Copyright DST Group. Licensed under the MIT license.
import enum


class TrinaryEnum(enum.Enum):
    TRUE = enum.auto()
    UNKNOWN = enum.auto()
    FALSE = enum.auto()

    @classmethod
    def parse_bool(cls, state_bool):
        if type(state_bool) is bool:
            if state_bool:
                return cls.TRUE
            else:
                return cls.FALSE
        else:
            return cls.UNKNOWN

    def __str__(self):
        return self.name

    def __eq__(self, other):
        if type(other) is bool:
            other = TrinaryEnum.parse_bool(other)
        if type(other) is not TrinaryEnum:
            return False
        if self.value == other.value:
            return True
        else:
            return False


class OperatingSystemPatch(enum.Enum):
    UNKNOWN = enum.auto()
    FILE_1 = enum.auto()
    Q147222 = enum.auto()
    KB911164 = enum.auto()
    MS17_010 = enum.auto()
    KB4500331 = enum.auto()
    KB4499149 = enum.auto()
    KB4499180 = enum.auto()
    KB4499164 = enum.auto()
    KB4499175 = enum.auto()


    @classmethod
    def parse_string(cls, patch_string):
        if patch_string.lower() == "file 1":
            return cls.FILE_1
        elif patch_string.lower() == "q147222":
            return cls.Q147222
        elif patch_string.lower() == "kb911164":
            return cls.KB911164
        elif patch_string.lower() == "ms17-010":
            return cls.MS17_010
        elif patch_string.lower() == "kb4500331":
            return cls.KB4500331
        elif patch_string.lower() == "kb4499149":
            return cls.KB4499149
        elif patch_string.lower() == "kb4499180":
            return cls.KB4499180
        elif patch_string.lower() == "kb4499164":
            return cls.KB4499164
        elif patch_string.lower() == "kb4499175":
            return cls.KB4499175

        return cls.UNKNOWN

    def __str__(self):
        return str(self.value)


class Architecture(enum.Enum):
    x86 = enum.auto()
    x64 = enum.auto()
    UNKNOWN = enum.auto()

    @classmethod
    def parse_string(cls, arch_string):
        if arch_string.lower() == "x86":
            return cls.x86
        if arch_string.lower() == "x64" or arch_string.lower() == "x86_64":
            return cls.x64
        else:
            return cls.UNKNOWN


class OperatingSystemType(enum.Enum):
    """An enum for representing the different possible Operating systems. """
    UNKNOWN = enum.auto()
    WINDOWS = enum.auto()
    LINUX = enum.auto()

    @classmethod
    def parse_string(cls, os_string):
        if os_string.lower() == "linux":
            return cls.LINUX
        elif "windows" in os_string.lower():
            return cls.WINDOWS
        else:
            return cls.UNKNOWN

    def __str__(self):
        return self.name


class OperatingSystemDistribution(enum.Enum):
    """An enum for representing the different possible Operating systems. """
    UNKNOWN = enum.auto()
    WINDOWS_XP = enum.auto()
    WINDOWS_SVR_2003 = enum.auto()
    WINDOWS_SVR_2008 = enum.auto()
    WINDOWS_7 = enum.auto()
    WINDOWS_8 = enum.auto()
    WINDOWS_10 = enum.auto()
    UBUNTU = enum.auto()
    KALI = enum.auto()
    WINDOWS_SVR_2003SP2 = enum.auto()
    WINDOWS_VISTA = enum.auto()
    WINDOWS_SVR_2008SP1 = enum.auto()
    WINDOWS_SVR_2008R2 = enum.auto()
    WINDOWS_7SP1 = enum.auto()

    @classmethod
    def parse_string(cls, os_string):
        os_string = os_string.lower()
        if os_string == "windows xp":
            return cls.WINDOWS_XP
        elif os_string == "windows server 2003":
            return cls.WINDOWS_SVR_2003
        elif "windows server 2008" in os_string or os_string == 'windows_svr_2008':
            return cls.WINDOWS_SVR_2008
        elif os_string == "windows 7":
            return cls.WINDOWS_7
        elif os_string == "windows 8":
            return cls.WINDOWS_8
        elif os_string == "windows 10":
            return cls.WINDOWS_10
        elif "ubuntu" in os_string:
            return cls.UBUNTU
        elif "kali" in os_string:
            return cls.KALI
        elif "2003sp2" in os_string and "window" in os_string:
            return cls.WINDOWS_SVR_2003SP2
        elif "vista" in os_string and "window" in os_string:
            return cls.WINDOWS_VISTA
        elif "svr 2008sp1" in os_string and "window" in os_string:
            return cls.WINDOWS_SVR_2008SP1
        elif "svr 2008r2" in os_string and "window" in os_string:
            return cls.WINDOWS_SVR_2008R2
        elif "7sp1" in os_string and "window" in os_string:
            return cls.WINDOWS_7SP1
        else:
            return cls.UNKNOWN

    def __str__(self):
        return self.name


class OperatingSystemVersion(enum.Enum):
    """An enum for representing the different possible Operating systems. """
    UNKNOWN = enum.auto()
    SP0 = enum.auto()
    SP1 = enum.auto()
    SP2 = enum.auto()
    SP3 = enum.auto()
    U18_04_3 = enum.auto()
    U18_04 = enum.auto()
    U8_04 = enum.auto()
    K2019_1 = enum.auto()
    K2019_2 = enum.auto()
    K2019_4 = enum.auto()
    W6_2_9200_16384 = enum.auto()
    W6_1_7601 = enum.auto()
    NT6_1 = enum.auto()

    @classmethod
    def parse_string(cls, os_string):
        os_string = os_string.lower()
        if os_string == "sp0":
            return cls.SP0
        elif os_string == "sp1":
            return cls.SP1
        elif os_string == "sp2":
            return cls.SP2
        elif os_string == "sp3":
            return cls.SP3
        elif os_string == "6.2.9200.16384":
            return cls.W6_2_9200_16384
        elif "6.1.7601" in os_string or os_string == 'w6_1_7601':
            return cls.W6_1_7601
        elif os_string == "18.04.3" or os_string == 'u18_04_3':
            return cls.U18_04_3
        elif os_string == "18.04":
            return cls.U18_04
        elif os_string == "8.04":
            return cls.U8_04
        elif os_string == "2019.1":
            return cls.K2019_1
        elif os_string == "2019.2":
            return cls.K2019_2
        elif "4.19.0-kali4" in os_string or os_string == "k2019_4":
            return cls.K2019_4
        elif os_string == "nt6.1":
            return cls.NT6_1
        else:
            return cls.UNKNOWN

    def __str__(self):
        return self.name


class OperatingSystemKernelVersion(enum.Enum):
    """An enum for representing the different possible Operating systems. """
    UNKNOWN = enum.auto()
    L2_6_24 = enum.auto()
    L4_15_0_1057_AWS = enum.auto()
    L5_2_0 = enum.auto()
    L5_3_0 = enum.auto()

    @classmethod
    def parse_string(cls, os_string):
        os_string = os_string.lower()
        if os_string == "linux 2.6.24":
            return cls.L2_6_24
        elif os_string == "4.15.0-1057-aws":
            return cls.L4_15_0_1057_AWS
        elif os_string == "linux 5.2.0":
            return cls.L5_2_0
        elif os_string == "linux 5.3.0":
            return cls.L5_3_0
        else:
            return cls.UNKNOWN

    def __str__(self):
        return self.name


class ProcessName(enum.Enum):
    UNKNOWN = enum.auto()
    SVCHOST = enum.auto()
    INIT = enum.auto()
    CRON = enum.auto()
    UDEVD = enum.auto()
    JSVC = enum.auto()
    SSHD = enum.auto()
    MYSQLD_SAFE = enum.auto()
    MYSQLD = enum.auto()
    SMBD = enum.auto()
    SMTP = enum.auto()
    FEMITTER = enum.auto()
    APACHE2 = enum.auto()
    EXPLORER = enum.auto()
    LSASS = enum.auto()
    WINLOGON = enum.auto()
    SMSS = enum.auto()
    SYSTEM = enum.auto()
    SYSTEM_IDLE_PROCESS = enum.auto()
    SERVICES = enum.auto()
    SHELL = enum.auto()
    TELNET = enum.auto()
    SLEEP = enum.auto()
    JAVA = enum.auto()
    PS = enum.auto()
    VELOCLIENT = enum.auto()
    POWERSHELL = enum.auto()
    CMD = enum.auto()

    @classmethod
    def parse_string(cls, name):
        name = name.lower()
        if name == "svchost" or name == "svchost.exe":
            return cls.SVCHOST
        elif name == "init":
            return cls.INIT
        elif name == "cron":
            return cls.CRON
        elif name == "udevd":
            return cls.UDEVD
        elif name == "jsvc":
            return cls.JSVC
        elif name == "sshd" or name == "sshd.exe":
            return cls.SSHD
        elif name == "mysqld_safe":
            return cls.MYSQLD_SAFE
        elif name == "mysqld":
            return cls.MYSQLD
        elif name == "smbd":
            return cls.SMBD
        elif name == "smtp":
            return cls.SMTP
        elif name == "femitter.exe":
            return cls.FEMITTER
        elif name == "apache2":
            return cls.APACHE2
        elif name == "explorer" or name == "explorer.exe":
            return cls.EXPLORER
        elif name == "lsass" or name == "lsass.exe":
            return cls.LSASS
        elif name == "winlogon" or name == "winlogon.exe":
            return cls.WINLOGON
        elif name == "smss" or name == "smss.exe":
            return cls.SMSS
        elif name == "system":
            return cls.SYSTEM
        elif name == "system idle process" or name == "system process":
            return cls.SYSTEM_IDLE_PROCESS
        elif name == "services" or name == "services.exe":
            return cls.SERVICES
        elif name == "bash" or name == "sh" or name == "sh.exe":
            return cls.SHELL
        elif name == "telnet":
            return cls.TELNET
        elif name == "sleep":
            return cls.SLEEP
        elif name == "java":
            return cls.JAVA
        elif name == "ps":
            return cls.PS
        elif name == "velociraptorclient":
            return cls.VELOCLIENT
        elif name == "powershell.exe" or name == "powershell":
            return cls.POWERSHELL
        elif name == "cmd.exe" or name == "cmd":
            return cls.CMD
        else:
            return cls.UNKNOWN


class ProcessType(enum.Enum):
    """An enum for representing the different types of services. """
    UNKNOWN = enum.auto()
    SSH = enum.auto()
    SVCHOST = enum.auto()
    SMB = enum.auto()
    SMTP = enum.auto()
    FEMITTER = enum.auto()
    WEBSERVER = enum.auto()
    NETCAT = enum.auto()
    RDP = enum.auto()
    REVERSE_SESSION_HANDLER = enum.auto()
    REVERSE_SESSION = enum.auto()
    MYSQL = enum.auto()

    @classmethod
    def parse_string(cls, service_string):
        service_string = service_string.lower()
        if service_string == "ssh" or service_string == "sshd" or service_string == "sshd.exe":
            return cls.SSH
        elif service_string == "svchost":
            return cls.SVCHOST
        elif service_string == "smtp":
            return cls.SMTP
        elif service_string == "femitter":
            return cls.FEMITTER
        elif service_string == "mysql":
            return cls.MYSQL
        elif service_string == "smb":
            return cls.SMB
        elif service_string.replace(" ", "") == "webserver":
            return cls.WEBSERVER
        elif service_string == "netcat":
            return cls.NETCAT
        elif service_string == "rdp":
            return cls.RDP
        elif service_string == "reverse_session_handler":
            return cls.REVERSE_SESSION_HANDLER
        elif service_string == "reverse_session":
            return cls.REVERSE_SESSION
        elif service_string == "http":
            return cls.WEBSERVER
        elif service_string == "https":
            return cls.WEBSERVER
        else:
            return cls.UNKNOWN

    def __str__(self):
        return self.name


# Potentially split this into separate Enums for each ProcessType at later date
class ProcessVersion(enum.Enum):
    OPENSSH_1_3 = enum.auto()
    SVC10_0_17763_1 = enum.auto()
    SAMBA_3_0_20_DEB = enum.auto()
    SMBv1 = enum.auto()
    APACHE_TOMCAT = enum.auto()
    PYTHON_SERVER = enum.auto()
    HARAKA_2_7_0 = enum.auto()
    HARAKA_2_8_9 = enum.auto()
    UNKNOWN = enum.auto()

    @classmethod
    def parse_string(cls, version_string):
        if version_string is not None and isinstance(version_string, str):
            version_string = version_string.lower()

        if version_string == "openssh 1.3":
            return cls.OPENSSH_1_3
        elif version_string == "10.0.17763.1":
            return cls.SVC10_0_17763_1
        elif version_string == "samba 3.0.20-debian":
            return cls.SAMBA_3_0_20_DEB
        elif version_string == "apache tomcat":
            return cls.APACHE_TOMCAT
        elif version_string == "python simplehttpserver":
            return cls.PYTHON_SERVER
        elif version_string == "smbv1":
            return cls.SMBv1
        elif version_string == "haraka 2.7.0":
            return cls.HARAKA_2_7_0
        elif version_string is not None:
            return version_string
        else:
            return cls.UNKNOWN


class TransportProtocol(enum.Enum):
    """An enum for representing the different types of services. """
    UNKNOWN = enum.auto()
    TCP = enum.auto()
    UDP = enum.auto()

    @classmethod
    def parse_string(cls, service_string):
        service_string = service_string.lower()
        if service_string == "tcp":
            return cls.TCP
        elif service_string == "udp":
            return cls.UDP
        else:
            return cls.UNKNOWN

    def __str__(self):
        return self.name


class BuiltInGroups(enum.Enum):
    UNKNOWN = enum.auto()
    USERS = enum.auto()
    WEBSERVER = enum.auto()
    ROOT = enum.auto()
    SHADOW = enum.auto()
    ADMINISTRATORS = enum.auto()

    @classmethod
    def parse_string(cls, group_string):
        group_string = group_string.lower()
        if group_string == "users":
            return cls.USERS
        elif group_string == "web server users":
            return cls.WEBSERVER
        elif group_string == "root":
            return cls.ROOT
        elif group_string == "shadow":
            return cls.SHADOW
        elif group_string == "administrators":
            return cls.ADMINISTRATORS
        else:
            return cls.UNKNOWN


class SessionType(enum.Enum):
    """An enum for representing the different types of sessions. """
    UNKNOWN = enum.auto()
    SSH = enum.auto()
    SHELL = enum.auto()
    METERPRETER = enum.auto()
    MSF_SHELL = enum.auto()
    MSF_SERVER = enum.auto()
    VELOCIRAPTOR_CLIENT = enum.auto()
    VELOCIRAPTOR_SERVER = enum.auto()
    LOCAL_SHELL = enum.auto()
    RED_ABSTRACT_SESSION = enum.auto()
    RED_REVERSE_SHELL = enum.auto()
    GREY_SESSION = enum.auto()

    @classmethod
    def parse_string(cls, service_string):
        service_string = service_string.lower()
        if service_string == "ssh":
            return cls.SSH
        elif service_string == "shell":
            return cls.SHELL
        elif service_string == "meterpreter":
            return cls.METERPRETER
        elif service_string == "msf shell" or service_string == "msf_shell":
            return cls.MSF_SHELL
        elif service_string == "metasploitserver":
            return cls.MSF_SERVER
        elif service_string == "velociraptorclient":
            return cls.VELOCIRAPTOR_CLIENT
        elif service_string == "velociraptorserver":
            return cls.VELOCIRAPTOR_SERVER
        elif service_string == "redabstractsession":
            return cls.RED_ABSTRACT_SESSION
        elif service_string == "red_reverse_shell":
            return cls.RED_REVERSE_SHELL
        elif service_string.replace(" ", "").replace("_", "") == "localshell":
            return cls.LOCAL_SHELL
        elif service_string == "green_session":
            return cls.GREY_SESSION
        else:
            return cls.UNKNOWN

    def __str__(self):
        return self.name


class Path(enum.Enum):
    UNKNOWN = enum.auto()
    WINDOWS = enum.auto()
    WINDOWS_SYSTEM = enum.auto()
    SYSTEM = enum.auto()
    TEMP = enum.auto()
    SBIN = enum.auto()
    BIN = enum.auto()
    USR_SBIN = enum.auto()
    USR_BIN = enum.auto()
    ETC = enum.auto()
    ADMINISTRATOR_DESKTOP = enum.auto()
    WEBSERVER = enum.auto()
    EXPLOIT = enum.auto()

    @classmethod
    def parse_string(cls, path_string):
        path_string = path_string.lower()
        if path_string == "system":
            return cls.SYSTEM
        elif path_string == "c:/windows/system32/":
            return cls.WINDOWS_SYSTEM
        elif path_string == "c:\\windows\\system32\\":
            return cls.WINDOWS_SYSTEM
        elif path_string == "c:/windows/" or path_string == "c:\\windows\\":
            return cls.WINDOWS
        elif path_string == "/tmp/" or path_string == "c:\\temp\\":
            return cls.TEMP
        elif path_string == "/sbin/" or path_string == "/sbin":
            return cls.SBIN
        elif path_string == "/bin/" or path_string == "/bin":
            return cls.BIN
        elif path_string == "/usr/sbin/" or path_string == "/usr/sbin":
            return cls.USR_SBIN
        elif path_string == "/usr/bin/" or path_string == "/usr/bin":
            return cls.USR_BIN
        elif path_string == "/etc/" or path_string == "/etc":
            return cls.ETC
        elif path_string == "c:\\users\\administrator\\desktop\\":
            return cls.ADMINISTRATOR_DESKTOP
        elif path_string == "/tmp/webserver/":
            return cls.WEBSERVER
        elif path_string == "/usr/share/exploitdb/exploits/linux/local/":
            return cls.EXPLOIT
        else:
            return cls.UNKNOWN


class ProcessState(enum.Enum):
    """An enum for representing the different types of services. """
    UNKNOWN = enum.auto()
    OPEN = enum.auto()
    CLOSED = enum.auto()
    FILTERED = enum.auto()

    @classmethod
    def parse_string(cls, service_string):
        service_string = service_string.lower()
        if service_string == "open":
            return cls.OPEN
        elif service_string == "closed":
            return cls.CLOSED
        elif service_string == "filtered":
            return cls.FILTERED
        else:
            return cls.UNKNOWN

    def __str__(self):
        return self.name


class FileType(enum.Enum):
    UNKNOWN = enum.auto()
    SVCHOST = enum.auto()
    PASSWD = enum.auto()
    SHADOW = enum.auto()
    FLAG = enum.auto()
    SMBCLIENT = enum.auto()
    NMAP = enum.auto()
    DirtyCowCode = enum.auto()
    DirtyCowExe = enum.auto()
    PYTHON = enum.auto()
    GCC = enum.auto()
    UDEV141CODE = enum.auto()
    UDEV141EXE = enum.auto()
    NC_REVERSE_SHELL = enum.auto()
    NC = enum.auto()

    @classmethod
    def parse_string(cls, name_string):
        name_string = name_string.lower()
        if name_string == "svchost":
            return cls.SVCHOST
        elif name_string == "passwd":
            return cls.PASSWD
        elif name_string == "shadow":
            return cls.SHADOW
        elif name_string == "flag":
            return cls.FLAG
        elif name_string == "smbclient":
            return cls.SMBCLIENT
        elif name_string == "nmap":
            return cls.NMAP
        elif name_string == "dirty_cow_c_file":
            return cls.DirtyCowCode
        elif name_string == "python":
            return cls.PYTHON
        elif name_string == "gcc":
            return cls.GCC
        elif name_string == "udev < 1.4.1":
            return cls.UDEV141CODE
        elif name_string == "nc_reverse_shell":
            return cls.NC_REVERSE_SHELL
        elif name_string == "nc":
            return cls.NC
        return cls.UNKNOWN


class FileVersion(enum.Enum):
    UNKNOWN = enum.auto()
    U4_2_4_1 = enum.auto()
    D9_2_1_21 = enum.auto()
    OPENBSD = enum.auto()

    @classmethod
    def parse_string(cls, name_string):
        name_string = name_string.lower()
        if name_string == "ubuntu 4.2.4-1":
            return cls.U4_2_4_1
        elif name_string == "debian 9.2.1-21":
            return cls.D9_2_1_21
        elif name_string == "openbsd":
            return cls.OPENBSD
        return cls.UNKNOWN


class FileExt(enum.Enum):
    ELF = enum.auto()
    UNKNOWN = enum.auto()

    @classmethod
    def parse_string(cls, name_string):
        if name_string.lower() == "elf":
            return cls.ELF
        return cls.UNKNOWN


class Vulnerability(enum.Enum):
    UNKNOWN = enum.auto()

    @classmethod
    def parse_string(cls, vuln_string):
        return cls.UNKNOWN


class Vendor(enum.Enum):
    UNKNOWN = enum.auto()

    @classmethod
    def parse_string(cls, vendor_string):
        return cls.UNKNOWN


class PasswordHashType(enum.Enum):
    UNKNOWN = enum.auto()
    MD5 = enum.auto()
    SHA512 = enum.auto()
    NTLM = enum.auto()

    @classmethod
    def parse_string(cls, hash_string):
        hash_string = hash_string.lower()
        if hash_string == "md5":
            return cls.MD5
        elif hash_string == 'sha512':
            return cls.SHA512
        elif hash_string == 'ntlm':
            return cls.NTLM
        return cls.UNKNOWN


class InterfaceType(enum.Enum):
    UNKNOWN = enum.auto()
    BROADCAST = enum.auto()
    LOCAL = enum.auto()

    @classmethod
    def parse_string(cls, interface_string):
        if interface_string.lower() == "broadcast":
            return cls.BROADCAST
        elif interface_string.lower() == "local":
            return cls.LOCAL
        return cls.UNKNOWN


class AppProtocol(enum.Enum):
    UNKNOWN = enum.auto()
    HTTP = enum.auto()
    HTTPS = enum.auto()
    SSH = enum.auto()
    JPV13 = enum.auto()
    TCP = enum.auto()
    MYSQL = enum.auto()
    NETBIOS_SSN = enum.auto()
    MICROSOFT_DS = enum.auto()
    RPC = enum.auto()

    @classmethod
    def parse_string(cls, protocol_string):
        protocol_string = protocol_string.lower()
        if protocol_string == "http":
            return cls.HTTP
        elif protocol_string == "https":
            return cls.HTTPS
        elif protocol_string == "ssh":
            return cls.SSH
        elif protocol_string == "jpv13":
            return cls.JPV13
        elif protocol_string == "tcp":
            return cls.TCP
        elif protocol_string == "mysql":
            return cls.MYSQL
        elif protocol_string == "netbios-ssn":
            return cls.NETBIOS_SSN
        elif protocol_string == "microsoft-ds":
            return cls.MICROSOFT_DS
        elif protocol_string == "rpc":
            return cls.RPC
        return cls.UNKNOWN


class QueryType(enum.Enum):
    SYNC = enum.auto()
    ASYNC = enum.auto()

    @classmethod
    def parse_string(cls, query_string):
        if query_string.lower() == "sync":
            return cls.SYNC
        elif query_string.lower() == "async":
            return cls.ASYNC

## The following code contains work of the United States Government and is not subject to domestic copyright protection under 17 USC § 105.
## Additionally, we waive copyright and related rights in the utilized code worldwide through the CC0 1.0 Universal public domain dedication.

class DecoyType(enum.Flag):
    NONE = 0
    ESCALATE = enum.auto()
    EXPLOIT = enum.auto()
    SANDBOXING_EXPLOIT = enum.auto()



# Copyright DST Group. Licensed under the MIT license.
from csle_cyborg.shared.actions.msf_actions_folder.meterpreter_actions_folder.meterpreter_action import MeterpreterAction
from csle_cyborg.shared.enums import SessionType
from csle_cyborg.shared.observation import Observation
from csle_cyborg.simulator.state import State


# Call reboot from a meterpreter session - reboots dict that session is on
# This deletes all non-default processes, and deletes all files in the /tmp/ folder
class MeterpreterReboot(MeterpreterAction):
    def __init__(self, session: int, agent: str, target_session: int):
        super().__init__(session=session, agent=agent, target_session=target_session)

    def sim_execute(self, state: State):
        obs = Observation()
        obs.set_success(False)
        if self.session not in state.sessions[self.agent]:
            return obs


        if self.session not in state.sessions[self.agent] or state.sessions[self.agent][
            self.session].session_type != SessionType.MSF_SERVER:
            return obs
        if self.meterpreter_session not in state.sessions[self.agent] or state.sessions[self.agent][
            self.meterpreter_session].session_type != SessionType.METERPRETER:
            return obs
        if state.sessions[self.agent][self.session].active and state.sessions[self.agent][self.meterpreter_session].active:
            host = state.sessions[self.agent][self.meterpreter_session].host
            obs.set_success(True)
            state.reboot_host(host)

        return obs



# Copyright DST Group. Licensed under the MIT license.
from ipaddress import IPv4Address, IPv4Network

from csle_cyborg.shared.actions.session_action import SessionAction
from csle_cyborg.simulator.interface import Interface
from csle_cyborg.simulator.msf_server_session import MSFServerSession
from csle_cyborg.simulator.session import Session
from csle_cyborg.simulator.state import State
from csle_cyborg.simulator.subnet import Subnet

lo_subnet = IPv4Network('127.0.0.0/8')
lo = IPv4Address('127.0.0.1')

class MSFAction(SessionAction):
    def __init__(self, session, agent):
        super().__init__(session)
        self.agent = agent

    def get_local_source_interface(self, local_session: Session, remote_address: IPv4Address, state: State) -> (Session, Interface):
        # discovers the local session and interface from routing through existing sessions
        remote_subnet = state.get_subnet_containing_ip_address(remote_address)
        # if MSF server then attempt to use the routes generated by autoroute
        if type(local_session) is MSFServerSession:
            for session, interfaces in local_session.routes.items():
                for interface in interfaces:
                    # find if remote address is in the sessions subnet
                    if remote_address in interface.subnet:
                        return local_session.children[session], interface
                    # find if the remote address is in a routable subnet
                    if interface.name in remote_subnet.nacls:
                        return local_session, interface
        for interface in state.hosts[local_session.host].interfaces:
            # find if remote address is in the sessions subnet
            if remote_address in interface.subnet:
                return local_session, interface

            # find if the remote address is in a routable subnet
            if interface.name != 'lo' and state.subnets[interface.subnet].name in remote_subnet.nacls:
                return local_session, interface

        return None, None

    def test_nacl(self, port, target_subnet: Subnet, originating_subnet: Subnet) -> bool:
        # return true if target subnet can receive traffic from originating subnet over specified port
        if originating_subnet == target_subnet:
            #no nacl block inside subnet
            return True

        if originating_subnet.name not in target_subnet.nacls:
            return False
        if 'all' in [i for i in target_subnet.nacls[originating_subnet.name]['in']]:
            return True
        if port in [i['PortRange'] for i in target_subnet.nacls[originating_subnet.name]['in'] if
                    type(i['PortRange']) is int]:
            return True
        return False

    def __str__(self):
        return f"{self.__class__.__name__}: MSF Session: {self.session}"



from csle_cyborg.agents.simple_agents.base_agent import BaseAgent
import random
from ipaddress import IPv4Network, IPv4Address
from csle_cyborg.shared.actions.shell_actions_folder.network_scan_folder.nmap_scan import NmapScan
from csle_cyborg.shared.actions.shell_actions_folder.network_scan_folder.ping_sweep import PingSweep
from csle_cyborg.shared.actions.msf_actions_folder.remote_code_execution_folder.samba_usermap_script import SambaUsermapScript
from csle_cyborg.shared.actions.msf_actions_folder.upgrade_to_meterpreter import UpgradeToMeterpreter
from csle_cyborg.shared.actions.msf_actions_folder.msf_auto_route import MSFAutoroute
from csle_cyborg.shared.actions.msf_actions_folder.remote_code_execution_folder.msf_eternal_blue import MSFEternalBlue
from csle_cyborg.shared.actions.msf_actions_folder.meterpreter_actions_folder.get_shell import GetShell
from csle_cyborg.shared.actions.shell_actions_folder.find_flag import FindFlag
from csle_cyborg.shared.actions.msf_actions_folder.msf_scanner_folder.tomcat_credential_scanner import TomcatCredentialScanner
from csle_cyborg.shared.actions.msf_actions_folder.remote_code_execution_folder.ssh_login_exploit import SSHLoginExploit
from csle_cyborg.shared.actions.msf_actions_folder.remote_code_execution_folder.tomcat_exploit import TomcatExploit


class KillchainAgent(BaseAgent):

    def __init__(self, action_size=None, state_size=None):

        self.colour = 'Red'
        self.killchains = {
            'smb': [NmapScan, SambaUsermapScript, UpgradeToMeterpreter, MSFAutoroute, PingSweep,
                    MSFEternalBlue, GetShell, FindFlag],
            'tomcat': [NmapScan, TomcatCredentialScanner, TomcatExploit, MSFAutoroute, PingSweep,
                       MSFEternalBlue, GetShell, FindFlag],
            'brute_force': [NmapScan, SSHLoginExploit, UpgradeToMeterpreter, MSFAutoroute, PingSweep,
                            MSFEternalBlue, GetShell, FindFlag],
        }

        self.action_parameters = {
            # 'Sleep':{},
            # 'IFConfig': {'agent': 'Red', 'session': 2},  # For Linux Box
            # 'IPConfig': {'agent': 'Red', 'session': 3},  # For Windows Box
            'FindFlag': {'agent': 'Red', 'session': 5},  # For Windows Box
            # 'SSHAccess':{'agent':'Red', 'session':3},  # From Linux to Windows
            'PingSweep': {'agent': 'Red', 'session': 2, 'subnet': IPv4Network('10.0.1.0/24')},  # From Linux to Windows
            'NmapScan': {'agent': 'Red', 'session': 1, 'subnet': IPv4Network('10.0.0.0/24')},  # From Kali to Linux
            # 'DeleteFileLinux':{'agent':'Red',# 'session':2},
            # 'DeleteFileWindows':{'agent':'Red', 'session':3},
            # 'KillProcessLinux':{'agent':'Red', 'session':2},
            # 'KillProcessWindows':{'agent':'Red', 'session':1},
            # 'AddUserWindows':{'agent':'Red', 'session':1},
            # 'AddUserLinux':{'agent':'Red', 'session':1},
            # 'DisableUserWindows':{'agent':'Red', 'session':1},
            # 'DisableUserLinux':{'agent':'Red', 'session':1},
            # 'RemoveUserFromGroupWindows':{'agent':'Red', 'session':1},
            # 'RemoveUserFromGroupLinux':{'agent':'Red', 'session':1},
            # 'DirtyCowPrivilegeEscalation':{'agent':'Red', 'session'},
            # 'LinuxKernelPrivilegeEscalation':{'agent':'Red', 'session':1},
            # 'SMBAnonymousConnection':{'agent':'Red', 'session':1},
            # 'ShellPS':{'agent':'Red', 'session':1},
            # 'SystemInfo':{'agent':'Red', 'session':1},
            # 'Uname':{'agent':'Red', 'session':1},
            # 'NetcatConnect':{'agent':'Red', 'session':1},
            # 'SSHHydraBruteForce':{'agent':'Red', 'session':1},
            # 'ReadShadowFile':{'agent':'Red', 'session':1},
            # 'ReadPasswdFile':{'agent':'Red', 'session':1},
            # 'StopService':{'agent':'Red', 'session':1},
            # 'StartService':{'agent':'Red', 'session':1},
            'EternalBlue': {'agent': 'Red', 'session': 0, 'ip_address':IPv4Address('10.0.1.2')},
            # 'GetPid':{'agent':'Red', 'session':1},
            'GetShell':{'agent':'Red', 'session':0, 'target_session':4},
            # 'GetUid':{'agent':'Red', 'session':1},
            # 'LocalTime':{'agent':'Red', 'session':1},
            'MSFAutoroute': {'agent': 'Red', 'session': 0, 'target_session': 3},
            # 'MeterpreterPS':{'agent':'Red', 'session':1},
            # 'MeterpreterReboot':{'agent':'Red', 'session':1},
            'SSHLoginExploit':{'agent':'Red', 'session':0, 'ip_address': IPv4Address('10.0.0.2'), 'port':22},
            'SambaUsermapScript': {'agent': 'Red', 'session': 0, 'ip_address': IPv4Address('10.0.0.2')},
            # 'Schtasks':{'agent':'Red', 'session':1},
            # 'SysInfo':{'agent':'Red', 'session':1},
            'TomcatCredentialScanner':{'agent':'Red', 'session':0, 'ip_address': IPv4Address('10.0.0.2'), 'port':8180},
            'TomcatExploit':{'agent':'Red', 'session':0, 'ip_address': IPv4Address('10.0.0.2'), 'port':8180,
                             'username':'tomcat', 'password':'tomcat'},
            'UpgradeToMeterpreter': {'agent': 'Red', 'session': 0, 'target_session': 2}
        }

        self.kchoice = random.choice(list(self.killchains.keys()))
        self.killchain = self.killchains[self.kchoice]
        if self.kchoice == 'tomcat':
            self.action_parameters['MSFAutoroute']['target_session'] -=1
            self.action_parameters['GetShell']['session'] -=1
            self.action_parameters['FindFlag']['session'] -= 1
        self.count = 0

    def train(self, results):
        pass

    def get_action(self, observation, action_space):
        position = self.count % len(self.killchain)
        action_class = self.killchain[position]
        action_params = self.action_parameters[action_class.__name__]
        self.count += 1
        return action_class(**action_params)

    def end_episode(self):
        pass

    def set_initial_values(self, action_space, observation):
        pass




# Copyright DST Group. Licensed under the MIT license.
import yaml
import pprint
from typing import List


class ScenarioSession:
    """A dataclass for handling scenario information of a session """

    def __init__(self, session_info: dict):
        """
        Parameters
        ----------
        session_info : dict
            Scenario definition of session
        """
        self._info = session_info

    @property
    def username(self) -> str:
        return self._info["username"]

    @property
    def session_type(self) -> str:
        return self._info["type"]

    @property
    def hostname(self) -> str:
        return self._info["hostname"]

    @property
    def parent(self) -> int:
        return self._info.get("parent", None)

    @property
    def num_children(self) -> int:
        return self._info.get("num_children_sessions", 0)

    @property
    def event_artifacts(self) -> List[str]:
        return self._info.get("artifacts", [])

    @property
    def name(self) -> str:
        return self._info.get("name", None)

    def __str__(self):
        return f"Session {self.name}, {self.parent} -> {self.username}@{self.hostname}: {self.session_type}"


class ScenarioAgent:
    """A dataclass for handling scenario information of an agent

    Is essentially a wrapper around the definition for a single agent
    in the scenario dictionary, and provides a consistent interface to
    agent data without having to remember string keys, etc.
    """

    def __init__(self, agent_name: str, agent_info: dict):
        """
        Parameters
        ----------
        agent_name: str
            Name of the agent
        agent_info : dict
            Scenario definition of the agent
        """
        self.name = agent_name
        self._info = agent_info
        self.starting_sessions = []
        for session_info in agent_info.get("starting_sessions", []):
            session = ScenarioSession(session_info)
            self.starting_sessions.append(session)

    @property
    def agent_type(self) -> str:
        return self._info.get("agent_type", None)

    @property
    def wrappers(self) -> List[str]:
        return self._info.get("wrappers", [])

    @property
    def actions(self) -> List[str]:
        return self._info.get("actions", [])

    @property
    def reward_calculator_type(self) -> str:
        return self._info.get("reward_calculator_type", None)

    @property
    def osint(self) -> dict:
        return self._info.get("INT", {})

    @property
    def allowed_subnets(self) -> List[str]:
        return self._info.get("AllowedSubnets", [])

    @property
    def adversary(self) -> str:
        return self._info.get('adversary')


class Scenario:
    """A dataclass for handling scenario information.

    Is essentially a wrapper around the scenario dictionary, that provides
    a consistent interface to scenario data.
    """

    def __init__(self, scenario_dict: dict):
        self._scenario = scenario_dict
        self._agents = {}
        for name, info in self._scenario["Agents"].items():
            self._agents[name] = ScenarioAgent(name, info)

    @classmethod
    def load(cls, scenario_file_path: str):
        with open(scenario_file_path) as fIn:
            scenario_dict = yaml.load(fIn, Loader=yaml.FullLoader)
        return cls(scenario_dict)

    @property
    def subnets(self) -> List[str]:
        return list(self._scenario["Subnets"])

    @property
    def hosts(self) -> List[str]:
        return list(self._scenario["Hosts"])

    @property
    def agents(self) -> List[str]:
        return list(self._agents)

    @property
    def starting_sessions(self) -> List[ScenarioSession]:
        sessions = []
        for agent in self._agents.values():
            for s in agent.starting_sessions:
                sessions.append(s)
        return sessions

    @property
    def predeployed(self) -> bool:
        return self._scenario.get("predeployed", False)

    def get_subnet_size(self, subnetname: str) -> int:
        return self._scenario["Subnets"][subnetname]["Size"]

    def get_subnet_hosts(self, subnetname: str) -> List[str]:
        return self._scenario["Subnets"][subnetname]["Hosts"]

    def get_subnet_nacls(self, subnetname: str) -> dict:
        subnet_info = self._scenario["Subnets"][subnetname]
        return subnet_info.get("NACLs", {})

    def get_host_image_name(self, hostname: str) -> str:
        return self._scenario["Hosts"][hostname]["image"]

    def get_host(self, hostname: str) -> dict:
        return self._scenario["Hosts"][hostname]

    def get_host_subnet_names(self, hostname: str) -> List[str]:
        host_subnets = []
        for subnetname in self.subnets:
            if hostname in self.get_subnet_hosts(subnetname):
                host_subnets.append(subnetname)
        return host_subnets

    def get_agent_info(self, agent_name: str) -> ScenarioAgent:
        return self._agents[agent_name]

    def __str__(self):
        return pprint.pformat(self._scenario, depth=2)
